import enum
from typing import Any, List, Optional, Sized, Tuple, Type, Union, Iterable

import numpy as np
import torch
from torch import nn

from ..base import SizedModule
from ...dimension import Dimension, DimensionProperty, DimensionsLike, SizeTypes
from ...transforms import to_tensor, ToDevice
from ...utils import format_pseudo_rn_seed, recursive_detach, ConnectivityConvention


class BaseLayer(SizedModule):
    """
    Base class for all layers.

    :Attributes:
        - :attr:`input_size` (Optional[Dimension]): The input size of the layer.
        - :attr:`output_size` (Optional[Dimension]): The output size of the layer.
        - :attr:`name` (str): The name of the layer.
        - :attr:`kwargs` (dict): Additional keyword arguments.

    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        """
        Constructor of the BaseLayer class.

        :param input_size: The input size of the layer.
        :type input_size: Optional[SizeTypes]
        :param output_size: The output size of the layer.
        :type output_size: Optional[SizeTypes]
        :param name: The name of the layer.
        :type name: Optional[str]
        :param device: The device of the layer. Defaults to the current available device.
        :type device: Optional[torch.device]
        :param kwargs: Additional keyword arguments.

        :keyword bool regularize: Whether to regularize the layer. If True, the method `update_regularization_loss`
            will be called after each forward pass. Defaults to False.
        :keyword bool freeze_weights: Whether to freeze the weights of the layer. Defaults to False.
        """
        super(BaseLayer, self).__init__(input_size=input_size, output_size=output_size, name=name)
        self._is_built = False
        self._freeze_weights = kwargs.get("freeze_weights", False)
        self._device = device
        if self._device is None:
            self._set_default_device_()
        self._device_transform = ToDevice(self.device)

        self.kwargs = kwargs
        self._set_default_kwargs()

        self.input_size = input_size
        self.output_size = output_size

        self._regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.device)

    @property
    def freeze_weights(self) -> bool:
        return self._freeze_weights

    @freeze_weights.setter
    def freeze_weights(self, freeze_weights: bool):
        self._freeze_weights = freeze_weights
        self.requires_grad_(self.requires_grad)

    @property
    def requires_grad(self):
        return not self.freeze_weights

    @property
    def is_ready_to_build(self) -> bool:
        return all(
            [
                s is not None
                for s in [
                self._input_size,
                (self._output_size if hasattr(self, "_output_size") else None)
            ]
            ]
        )

    @property
    def is_built(self) -> bool:
        return self._is_built

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device: torch.device):
        """
        Set the device of the layer and move all the parameters to the new device.

        :param device: The device to set.
        :type device: torch.device

        :return: None
        """
        self.to(device, non_blocking=True)

    def to(self, device: torch.device, non_blocking: bool = True, *args, **kwargs):
        """
        Move all the parameters of the layer to the specified device.

        :param device: The device to move the parameters to.
        :type device: torch.device
        :param non_blocking: Whether to move the parameters in a non-blocking way.
        :type non_blocking: bool
        :param args: Additional positional arguments.
        :param kwargs: Additional keyword arguments.

        :return: self
        """
        self._device = device
        for module in self.modules():
            if module is not self and getattr(module, "device", device).type != device.type:
                module.to(device, non_blocking=non_blocking)
        return super(BaseLayer, self).to(*args, **kwargs)

    def __repr__(self):
        _repr = f"{self.__class__.__name__}"
        if self.name_is_set:
            _repr += f"<{self.name}>"
        _repr += f"({int(self.input_size)}->{int(self.output_size)}"
        _repr += f"{self.extra_repr()})"
        _repr += f"@{self.device}"
        return _repr

    def _format_size(self, size: Optional[SizeTypes], **kwargs) -> Optional[Dimension]:
        kwargs["filter_time"] = True
        return super(BaseLayer, self)._format_size(size, **kwargs)

    def _set_default_kwargs(self):
        pass

    def _set_default_device_(self):
        self._device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    def build(self) -> 'BaseLayer':
        """
        Build the layer. This method must be call after the layer is initialized to make sure that the layer is ready
        to be used e.g. the input and output size is set, the weights are initialized, etc.

        :return: The layer itself.
        :rtype: BaseLayer
        """
        if self._is_built:
            raise ValueError("The layer can't be built multiple times.")
        if not self.is_ready_to_build:
            raise ValueError("Input size and output size must be specified before the build call.")
        self._is_built = True
        self.reset_regularization_loss()
        return self

    def create_empty_state(self, batch_size: int = 1, **kwargs) -> Tuple[torch.Tensor, ...]:
        """
        Create an empty state for the layer. This method must be implemented by the child class.

        :param batch_size: The batch size of the state.
        :type batch_size: int

        :return: The empty state.
        :rtype: Tuple[torch.Tensor, ...]
        """
        raise NotImplementedError()

    def _init_forward_state(
            self,
            state: Tuple[torch.Tensor, ...] = None,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        if state is None:
            state = self.create_empty_state(batch_size, **kwargs)
        elif isinstance(state, (list, tuple)) and any([e is None for e in state]):
            empty_state = self.create_empty_state(batch_size, **kwargs)
            state = list(state)
            for i, e in enumerate(state):
                if e is None:
                    state[i] = empty_state[i]
            state = tuple(state)
        return state

    def infer_sizes_from_inputs(self, inputs: torch.Tensor):
        """
        Try to infer the input and output size of the layer from the inputs.

        :param inputs: The inputs to infer the size from.
        :type inputs: torch.Tensor

        :return: None
        """
        self.input_size = inputs.shape[-1]
        if self.output_size is None:
            raise ValueError("output_size must be specified before the forward call.")

    def __call__(self, inputs: torch.Tensor, state: torch.Tensor = None, *args, **kwargs):
        """
        Call the forward method of the layer. If the layer is not built, it will be built automatically.
        In addition, if :attr:`kwargs['regularize']` is set to True, the :meth: `update_regularization_loss` method
        will be called.

        :param inputs: The inputs to the layer.
        :type inputs: torch.Tensor

        :param args: The positional arguments to the forward method.
        :param kwargs: The keyword arguments to the forward method.

        :return: The output of the layer.
        """
        inputs, state = self._device_transform(inputs), self._device_transform(state)
        if not self.is_built:
            if not self.is_ready_to_build:
                self.infer_sizes_from_inputs(inputs)
            self.build()
        call_output = super(BaseLayer, self).__call__(inputs, state, *args, **kwargs)

        if isinstance(call_output, torch.Tensor):
            hidden_state = None
        elif isinstance(call_output, (List, Tuple)) and len(call_output) == 2:
            hidden_state = call_output[1]
        else:
            raise ValueError(
                "The forward method must return a torch.Tensor (the output of the layer) "
                "or a tuple of torch.Tensor (the output of the layer and the hidden state)."
            )
        if self.kwargs.get("regularize", False):
            self.update_regularization_loss(hidden_state)
        return call_output

    def forward(
            self,
            inputs: torch.Tensor,
            state: torch.Tensor = None,
            **kwargs
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        raise NotImplementedError()

    def initialize_weights_(self):
        """
        Initialize the weights of the layer. This method must be implemented by the child class.

        :return: None
        """
        pass

    def update_regularization_loss(self, state: Optional[Any] = None, *args, **kwargs) -> torch.Tensor:
        """
        Update the regularization loss for this layer. Each update call increments the regularization loss so at the end
        the regularization loss will be the sum of all calls to this function. This method is called at the end of each
        forward call automatically by the BaseLayer class.

        :param state: The current state of the layer.
        :type state: Optional[Any]
        :param args: Other positional arguments.
        :param kwargs: Other keyword arguments.

        :return: The updated regularization loss.
        :rtype: torch.Tensor
        """
        return self._regularization_loss

    def reset_regularization_loss(self):
        """
        Reset the regularization loss to zero.

        :return: None
        """
        self._regularization_loss = torch.tensor(0.0, dtype=torch.float32, device=self.device)

    def get_and_reset_regularization_loss(self):
        """
        Get and reset the regularization loss for this layer. The regularization loss will be reset by the
        reset_regularization_loss method after it is returned.

        WARNING: If this method is not called after an integration, the update of the regularization loss can cause a
        memory leak. TODO: fix this.

        :return: The regularization loss.
        """
        loss = self.get_regularization_loss()
        self.reset_regularization_loss()
        return loss

    def get_regularization_loss(self) -> torch.Tensor:
        """
        Get the regularization loss for this layer.

        :return: The regularization loss.
        """
        return self._regularization_loss


# @inherit_fields_docstring(fields=["Attributes"], bases=[BaseLayer])
class BaseNeuronsLayer(BaseLayer):
    """
    A base class for layers that have neurons. This class provides two importants Parameters: the
    :attr:`forward_weights` and the :attr:`recurrent_weights`. Child classes must implement the :method:`forward`
    method and the :mth:`create_empty_state` method.

    :Attributes:
        - :attr:`forward_weights` (torch.nn.Parameter): The weights used to compute the output of the layer.
        - :attr:`recurrent_weights` (torch.nn.Parameter): The weights used to compute the hidden state of the layer.
        - :attr:`dt` (float): The time step of the layer.
        - :attr:`use_rec_eye_mask` (torch.Tensor): Whether to use the recurrent eye mask.
        - :attr:`rec_mask` (torch.Tensor): The recurrent eye mask.
    """

    def __init__(
            self,
            input_size: Optional[SizeTypes] = None,
            output_size: Optional[SizeTypes] = None,
            name: Optional[str] = None,
            use_recurrent_connection: bool = True,
            use_rec_eye_mask: bool = False,
            dt: float = 1e-3,
            device: Optional[torch.device] = None,
            **kwargs
    ):
        """
        Initialize the layer.; See the :class:`BaseLayer` class for more details.;

        :param input_size: The input size of the layer;
        :type input_size: Optional[SizeTypes]
        :param output_size: The output size of the layer.
        :type output_size: Optional[SizeTypes]
        :param name: The name of the layer.
        :type name: Optional[str]
        :param use_recurrent_connection: Whether to use a recurrent connection. Default is True.
        :type use_recurrent_connection: bool
        :param use_rec_eye_mask: Whether to use a recurrent eye mask. Default is False. This mask will be used to
            mask to zero the diagonal of the recurrent connection matrix.
        :type use_rec_eye_mask: bool
        :param dt: The time step of the layer. Default is 1e-3.
        :type dt: float
        :param kwargs: Other keyword arguments.

        :keyword bool regularize: Whether to regularize the layer. If True, the method `update_regularization_loss` will
            be called after each forward pass. Defaults to False.
        :keyword str hh_init: The initialization method for the hidden state. Defaults to "zeros".
        :keyword float hh_init_mu: The mean of the hidden state initialization when hh_init is random . Defaults to 0.0.
        :keyword float hh_init_std: The standard deviation of the hidden state initialization when hh_init is random. Defaults to 1.0.
        :keyword int hh_init_seed: The seed of the hidden state initialization when hh_init is random. Defaults to 0.
        :keyword bool force_dale_law: Whether to force the Dale's law in the layer's weights. Defaults to False.
        :keyword Union[torch.Tensor, float] forward_sign: If force_dale_law is True, this parameter will be used to
            initialize the forward_sign vector. If it is a float, the forward_sign vector will be initialized with this
            value as the ration of inhibitory neurons. If it is a tensor, it will be used as the forward_sign vector.
        :keyword Union[torch.Tensor, float] recurrent_sign: If force_dale_law is True, this parameter will be used to
            initialize the recurrent_sign vector. If it is a float, the recurrent_sign vector will be initialized with
            this value as the ration of inhibitory neurons. If it is a tensor, it will be used as the recurrent_sign vector.
        :keyword Callable sign_activation: The activation function used to compute the sign of the weights i.e. the
            forward_sign and recurrent_sign vectors. Defaults to torch.nn.Tanh.
        """
        self.dt = dt
        self.use_recurrent_connection = use_recurrent_connection
        self._forward_weights = None
        self._forward_sign = None
        self.use_rec_eye_mask = use_rec_eye_mask
        self._recurrent_weights = None
        self._recurrent_sign = None
        self.rec_mask = None
        self._force_dale_law = kwargs.get("force_dale_law", False)
        self._connectivity_convention = ConnectivityConvention.from_other(
            kwargs.get("connectivity_convention", ConnectivityConvention.ItoJ)
        )
        super().__init__(
            input_size=input_size,
            output_size=output_size,
            name=name,
            device=device,
            **kwargs
        )
        self.sign_activation = self.kwargs.get("sign_activation", torch.nn.Tanh())
        self.activation = self._init_activation(self.kwargs.get("activation", "identity"))

    @property
    def forward_weights(self) -> torch.nn.Parameter:
        """
        Get the forward weights.

        :return: The forward weights.
        """
        if self.force_dale_law:
            return torch.pow(self._forward_weights, 2) * self.forward_sign
        return self._forward_weights

    @forward_weights.setter
    def forward_weights(self, value: torch.nn.Parameter):
        """
        Set the forward weights.

        :param value: The forward weights.
        """
        if not isinstance(value, torch.nn.Parameter):
            value = torch.nn.Parameter(to_tensor(value), requires_grad=self.requires_grad)
        self._forward_weights = value

    @property
    def recurrent_weights(self) -> torch.nn.Parameter:
        """
        Get the recurrent weights.

        :return: The recurrent weights.
        """
        if self.force_dale_law:
            return torch.pow(self._recurrent_weights, 2) * self.recurrent_sign
        return self._recurrent_weights

    @recurrent_weights.setter
    def recurrent_weights(self, value: torch.nn.Parameter):
        """
        Set the recurrent weights.

        :param value: The recurrent weights.
        """
        if not isinstance(value, torch.nn.Parameter):
            value = torch.nn.Parameter(to_tensor(value), requires_grad=self.requires_grad)
        self._recurrent_weights = value

    @property
    def force_dale_law(self) -> bool:
        """
        Get whether to force the Dale's law.

        :return: Whether to force the Dale's law.
        """
        return self._force_dale_law

    @property
    def forward_sign(self) -> Optional[torch.nn.Parameter]:
        """
        Get the forward sign.

        :return: The forward sign.
        """
        if self._forward_sign is None:
            return None
        return self.sign_activation(self._forward_sign)

    @forward_sign.setter
    def forward_sign(self, value: torch.nn.Parameter):
        """
        Set the forward sign.

        :param value: The forward sign.
        """
        if not isinstance(value, torch.nn.Parameter):
            value = torch.nn.Parameter(value, requires_grad=self.force_dale_law and self.requires_grad)
        self._forward_sign = value

    @property
    def recurrent_sign(self) -> Optional[torch.nn.Parameter]:
        """
        Get the recurrent sign.

        :return: The recurrent sign.
        """
        if self._recurrent_sign is None:
            return None
        return self.sign_activation(self._recurrent_sign)

    @recurrent_sign.setter
    def recurrent_sign(self, value: torch.nn.Parameter):
        """
        Set the recurrent sign.

        :param value: The recurrent sign.
        """
        if not isinstance(value, torch.nn.Parameter):
            value = torch.nn.Parameter(value, requires_grad=self.force_dale_law)
        self._recurrent_sign = value

    @property
    def connectivity_convention(self) -> ConnectivityConvention:
        """
        Get the connectivity convention.

        :return: The connectivity convention.
        """
        return self._connectivity_convention

    def get_forward_weights_parameter(self) -> torch.nn.Parameter:
        """
        Get the forward weights parameter.

        :return: The forward weights parameter.
        """
        return self._forward_weights

    def set_forward_weights_parameter(self, parameter: torch.nn.Parameter):
        """
        Set the forward weights parameter.

        :param parameter: The forward weights parameter.
        """
        self._forward_weights = parameter

    def get_recurrent_weights_parameter(self) -> torch.nn.Parameter:
        """
        Get the recurrent weights parameter.

        :return: The recurrent weights parameter.
        """
        return self._recurrent_weights

    def set_recurrent_weights_parameter(self, parameter: torch.nn.Parameter):
        """
        Set the recurrent weights parameter.

        :param parameter: The recurrent weights parameter.
        """
        self._recurrent_weights = parameter

    def get_forward_sign_parameter(self) -> torch.nn.Parameter:
        """
        Get the forward sign parameter.

        :return: The forward sign parameter.
        """
        return self._forward_sign

    def set_forward_sign_parameter(self, parameter: torch.nn.Parameter):
        """
        Set the forward sign parameter.

        :param parameter: The forward sign parameter.
        """
        self._forward_sign = parameter

    def get_recurrent_sign_parameter(self) -> torch.nn.Parameter:
        """
        Get the recurrent sign parameter.

        :return: The recurrent sign parameter.
        """
        return self._recurrent_sign

    def set_recurrent_sign_parameter(self, parameter: torch.nn.Parameter):
        """
        Set the recurrent sign parameter.

        :param parameter: The recurrent sign parameter.
        """
        self._recurrent_sign = parameter

    def get_forward_weights_data(self) -> torch.Tensor:
        """
        Get the forward weights data.

        :return: The forward weights data.
        """
        return self._forward_weights.data

    def set_forward_weights_data(self, data: torch.Tensor):
        """
        Set the forward weights data.

        :param data: The forward weights data.
        """
        self._forward_weights.data = to_tensor(data).to(self.device)

    def get_recurrent_weights_data(self) -> torch.Tensor:
        """
        Get the recurrent weights data.

        :return: The recurrent weights data.
        """
        return self._recurrent_weights.data

    def set_recurrent_weights_data(self, data: torch.Tensor):
        """
        Set the recurrent weights data.

        :param data: The recurrent weights data.
        """
        self._recurrent_weights.data = to_tensor(data).to(self.device)

    def get_weights_parameters(self) -> List[torch.nn.Parameter]:
        """
        Get the weights parameters.

        :return: The weights parameters.
        """
        parameters = [self._forward_weights]
        if self.use_recurrent_connection:
            parameters.append(self._recurrent_weights)
        return parameters

    def get_sign_parameters(self) -> List[torch.nn.Parameter]:
        """
        Get the sign parameters.

        :return: The sign parameters.
        """
        parameters = []
        if self.force_dale_law:
            parameters.append(self._forward_sign)
            if self.use_recurrent_connection:
                parameters.append(self._recurrent_sign)
        return parameters

    def create_empty_state(
            self,
            batch_size: int = 1,
            **kwargs
    ) -> Tuple[torch.Tensor, ...]:
        self.kwargs.setdefault("hh_init", "zeros")
        self.kwargs.setdefault("hh_init_mu", 0.0)
        self.kwargs.setdefault("hh_init_std", 1.0)

        n_hh = kwargs.get("n_hh", 1)
        if self.kwargs["hh_init"] == "zeros":
            state = tuple(
                [torch.zeros(
                    (batch_size, int(self.output_size)),
                    device=self.device,
                    dtype=torch.float32,
                    requires_grad=True,
                ) for _ in range(n_hh)]
            )
        elif self.kwargs["hh_init"] == "random":
            mu, std = self.kwargs.get("hh_init_mu", 0.0), self.kwargs.get("hh_init_std", 1.0)
            gen = torch.Generator(device=self.device)
            gen.manual_seed(format_pseudo_rn_seed(self.kwargs.get("hh_init_seed", None)))
            state = [(torch.rand(
                (batch_size, int(self.output_size)),
                device=self.device,
                dtype=torch.float32,
                requires_grad=True,
                generator=gen,
            ) * std + mu) for _ in range(n_hh)]
        elif self.kwargs["hh_init"] == "inputs":
            assert "inputs" in kwargs, "inputs must be provided to initialize the state"
            assert kwargs["inputs"].shape == (batch_size, int(self.output_size))
            state = [kwargs["inputs"].clone() for _ in range(n_hh)]
        elif self.kwargs["hh_init"].lower() == "given":
            assert "h0" in self.kwargs, "h0 must be provided as a tuple of tensors when hh_init is 'given'."
            h0 = self.kwargs["h0"]
            assert isinstance(h0, (tuple, list)), "h0 must be a tuple of tensors."
            state = [to_tensor(h0_, dtype=torch.float32).to(self.device) for h0_ in h0]
        else:
            raise ValueError("Hidden state init method not known. Please use 'zeros', 'inputs', 'random' or 'given'.")
        return tuple(state)

    def forward(
            self,
            inputs: torch.Tensor,
            state: torch.Tensor = None,
            **kwargs
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        raise NotImplementedError()

    def _init_forward_sign_(self):
        if self.kwargs.get("forward_sign", None) is None:
            self.kwargs.pop("forward_sign", None)
        if "forward_sign" in self.kwargs and self.force_dale_law:
            if isinstance(self.kwargs["forward_sign"], float):
                assert 0.0 <= self.kwargs["forward_sign"] <= 1.0, "forward_sign must be in [0, 1]"
                n_inh = int(int(self.input_size) * self.kwargs["forward_sign"])
                inh_indexes = torch.randperm(int(self.input_size))[:n_inh]
                self.kwargs["forward_sign"] = np.abs(np.random.normal(size=(int(self.input_size), 1)))
                self.kwargs["forward_sign"][inh_indexes] *= -1
            assert self.kwargs["forward_sign"].shape == (int(self.input_size), 1), \
                "forward_sign must be a float or a tensor of shape (input_size, 1)"
            self._forward_sign.data = to_tensor(self.kwargs["forward_sign"]).to(self.device)
            with torch.no_grad():
                self._forward_weights.data = torch.sqrt(torch.abs(self._forward_weights.data))
        elif self.force_dale_law:
            torch.nn.init.normal_(self._forward_sign)

    def _init_recurrent_sign_(self):
        if self.kwargs.get("recurrent_sign", None) is None:
            self.kwargs.pop("recurrent_sign", None)
        if "recurrent_sign" in self.kwargs and self.force_dale_law and self.use_recurrent_connection:
            if isinstance(self.kwargs["recurrent_sign"], float):
                assert 0.0 <= self.kwargs["recurrent_sign"] <= 1.0, "recurrent_sign must be in [0, 1]"
                n_inh = int(int(self.output_size) * self.kwargs["recurrent_sign"])
                inh_indexes = torch.randperm(int(self.output_size))[:n_inh]
                self.kwargs["recurrent_sign"] = np.abs(np.random.normal(size=(int(self.output_size), 1)))
                self.kwargs["recurrent_sign"][inh_indexes] *= -1
            assert self.kwargs["recurrent_sign"].shape == (int(self.output_size), 1), \
                "recurrent_sign must be a float or a tensor of shape (output_size, 1)"
            self._recurrent_sign.data = to_tensor(self.kwargs["recurrent_sign"]).to(self.device)
            with torch.no_grad():
                self._recurrent_weights.data = torch.sqrt(torch.abs(self._recurrent_weights.data))
        elif self.force_dale_law and self.use_recurrent_connection:
            torch.nn.init.xavier_normal_(self._recurrent_sign)

    def _init_activation(self, activation: Union[torch.nn.Module, str] = "identity"):
        """
        Initialise the activation function.

        :param activation: Activation function.
        :type activation: Union[torch.nn.Module, str]
        """
        str_to_activation = {
            "identity": torch.nn.Identity(),
            "relu"    : torch.nn.ReLU(),
            "tanh"    : torch.nn.Tanh(),
            "sigmoid" : torch.nn.Sigmoid(),
            "softmax" : torch.nn.Softmax(dim=-1),
            "elu"     : torch.nn.ELU(),
            "selu"    : torch.nn.SELU(),
            "prelu"   : torch.nn.PReLU(),
            "leakyrelu": torch.nn.LeakyReLU(),
            "leaky_relu": torch.nn.LeakyReLU(),
            "logsigmoid": torch.nn.LogSigmoid(),
            "log_sigmoid": torch.nn.LogSigmoid(),
            "logsoftmax": torch.nn.LogSoftmax(dim=-1),
            "log_softmax": torch.nn.LogSoftmax(dim=-1),
        }
        if isinstance(activation, str):
            if activation.lower() not in str_to_activation.keys():
                raise ValueError(
                    f"Activation {activation} is not implemented. Please use one of the following: "
                    f"{str_to_activation.keys()} or provide a torch.nn.Module."
                )
            self.activation = str_to_activation[activation.lower()]
        else:
            self.activation = activation
        return self.activation

    def initialize_weights_(self):
        super().initialize_weights_()
        if self.kwargs.get("forward_weights", None) is not None:
            self._forward_weights.data = to_tensor(self.kwargs["forward_weights"]).to(self.device)
        else:
            torch.nn.init.xavier_normal_(self._forward_weights)

        if self.kwargs.get("recurrent_weights", None) is not None and self.use_recurrent_connection:
            self._recurrent_weights.data = to_tensor(self.kwargs["recurrent_weights"]).to(self.device)
        elif self.use_recurrent_connection:
            torch.nn.init.xavier_normal_(self._recurrent_weights)

        self._init_forward_sign_()
        self._init_recurrent_sign_()

    def build(self) -> 'BaseNeuronsLayer':
        """
        Build the layer. This method must be call after the layer is initialized to make sure that the layer is ready
        to be used e.g. the input and output size is set, the weights are initialized, etc.

        In this method the :attr:`forward_weights`, :attr:`recurrent_weights` and :attr: `rec_mask` are created and
        finally the method :meth:`initialize_weights_` is called.

        :return: The layer itself.
        :rtype: BaseLayer
        """
        super().build()
        self._forward_weights = nn.Parameter(
            torch.empty((int(self.input_size), int(self.output_size)), device=self.device, dtype=torch.float32),
            requires_grad=self.requires_grad
        )
        if self.force_dale_law:
            self._forward_sign = torch.nn.Parameter(
                torch.empty((int(self.input_size), 1), dtype=torch.float32, device=self.device),
                requires_grad=self.force_dale_law
            )

        if self.use_recurrent_connection:
            self._recurrent_weights = nn.Parameter(
                torch.empty((int(self.output_size), int(self.output_size)), device=self.device, dtype=torch.float32),
                requires_grad=self.requires_grad
            )
            if self.use_rec_eye_mask:
                self.rec_mask = nn.Parameter(
                    (1 - torch.eye(int(self.output_size), device=self.device, dtype=torch.float32)),
                    requires_grad=False
                )
            else:
                self.rec_mask = nn.Parameter(
                    torch.ones(
                        (int(self.output_size), int(self.output_size)), device=self.device, dtype=torch.float32
                    ),
                    requires_grad=False
                )
            if self.force_dale_law:
                self._recurrent_sign = torch.nn.Parameter(
                    torch.empty((int(self.output_size), 1), dtype=torch.float32, device=self.device),
                    requires_grad=self.force_dale_law
                )
        self.initialize_weights_()
        return self

    def __repr__(self):
        _repr = f"{self.__class__.__name__}"
        if self.name_is_set:
            _repr += f"<{self.name}>"
        if self.force_dale_law:
            _repr += f"[Dale]"
        _repr += f"({int(self.input_size)}"
        if self.use_recurrent_connection:
            _repr += "<"
        _repr += f"->{int(self.output_size)}"
        _repr += f"{self.extra_repr()}"
        _repr += f", activation:{self.activation}"
        if self.force_dale_law:
            _repr += f", sign_activation:{self.sign_activation}"
        _repr += ")"
        if self.freeze_weights:
            _repr += "[frozen]"
        _repr += f"@{self.device}"
        return _repr





from .base import (
    to_tensor,
    to_numpy,
    LinearRateToSpikes,
    ConstantValuesTransform,
    IdentityTransform,
    ToDevice,
    ToTensor,
    ReduceMax,
    ReduceMean,
    ReduceSum,
    ReduceFuncTanh,
)

# from .spikes_encoders import (
# 	SpikesEncoder,
# 	LIFEncoder,
# 	SpyLIFLayer,
# 	ALIFEncoder,
# )






import numbers
from typing import Any, Optional, Callable, Union, Dict

import numpy as np
import torch


def to_tensor(x: Any, dtype=torch.float32):
    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).type(dtype)
    elif isinstance(x, torch.Tensor):
        return x.type(dtype)
    elif isinstance(x, numbers.Number):
        return torch.tensor(x, dtype=dtype)
    elif isinstance(x, dict):
        return {k: to_tensor(v, dtype=dtype) for k, v in x.items()}
    elif isinstance(x, (list, tuple)):
        return type(x)([to_tensor(v, dtype=dtype) for v in x])
    elif not isinstance(x, torch.Tensor):
        try:
            return torch.tensor(x, dtype=dtype)
        except Exception as e:
            raise ValueError(f"Unsupported type {type(x)}") from e
    raise ValueError(f"Unsupported type {type(x)}")


def to_numpy(x: Any, dtype=np.float32):
    if isinstance(x, np.ndarray):
        return np.asarray(x, dtype=dtype)
    elif isinstance(x, torch.Tensor):
        return x.detach().cpu().numpy()
    elif isinstance(x, numbers.Number):
        return x
    elif isinstance(x, dict):
        return {k: to_numpy(v, dtype=dtype) for k, v in x.items()}
    elif not isinstance(x, torch.Tensor):
        try:
            return np.asarray(x, dtype=dtype)
        except Exception as e:
            raise ValueError(f"Unsupported type {type(x)}") from e
    raise ValueError(f"Unsupported type {type(x)}")


class ToDevice(torch.nn.Module):
    def __init__(self, device: torch.device, non_blocking: bool = True):
        super().__init__()
        self._device = device
        self.non_blocking = non_blocking
        self.to(device)

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, device: torch.device):
        self._device = device
        self.to(device)

    def to(self, device: Optional[Union[int, torch.device]], non_blocking: Optional[bool] = None, *args, **kwargs):
        self._device = device
        if non_blocking is None:
            non_blocking = self.non_blocking
        else:
            self.non_blocking = non_blocking
        return super().to(device=device, non_blocking=non_blocking, *args, **kwargs)

    def forward(self, x: torch.Tensor):
        if x is None:
            return x
        if not isinstance(x, torch.Tensor):
            if isinstance(x, dict):
                return {k: self.forward(v) for k, v in x.items()}
            elif isinstance(x, list):
                return [self.forward(v) for v in x]
            elif isinstance(x, tuple):
                return tuple(self.forward(v) for v in x)
            else:
                return x
        return x.to(self.device, non_blocking=self.non_blocking)

    def __repr__(self):
        return f"ToDevice({self.device}, async={self.non_blocking})"


class ToTensor(torch.nn.Module):
    def __init__(self, dtype=torch.float32, device: Optional[torch.device] = None):
        super().__init__()
        self.dtype = dtype
        self.device = device
        if self.device is None:
            self.to_device = None
        else:
            self.to_device = ToDevice(self.device)

    def forward(self, x: Any) -> torch.Tensor:
        x = to_tensor(x, self.dtype)
        if self.to_device is not None:
            x = self.to_device(x)
        return x


class IdentityTransform(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: Any):
        return x


class LinearRateToSpikes(torch.nn.Module):
    def __init__(
            self,
            n_steps: int,
            *,
            data_min: float = 0.0,
            data_max: float = 1.0,
            epsilon: float = 1e-6,
    ):
        super().__init__()
        self.n_steps = n_steps
        self.data_min = to_tensor(data_min)
        self.data_max = to_tensor(data_max)
        self.epsilon = epsilon

    def __repr__(self):
        return f"{self.__class__.__name__}(n_steps={self.n_steps})"

    def firing_periods_to_spikes(self, firing_periods: np.ndarray) -> np.ndarray:
        firing_periods = np.floor(firing_periods).astype(int)
        ones_mask = firing_periods == 1
        firing_periods[firing_periods > self.n_steps] = self.n_steps
        firing_periods[firing_periods < 1] = 1
        time_dim = np.expand_dims(np.arange(1, self.n_steps+1), axis=tuple(np.arange(firing_periods.ndim)+1))
        spikes = (time_dim % firing_periods) == 0
        spikes[0, ones_mask] = 0
        return spikes.astype(float)

    def forward(self, x):
        x = to_tensor(x)
        device = x.device
        squeeze_flag = False
        if x.dim() < 1:
            x = x.unsqueeze(0)
            squeeze_flag = True
        self.data_min = torch.minimum(x, self.data_min)
        self.data_max = torch.maximum(x, self.data_max)
        x = (x - self.data_min) / (self.data_max - self.data_min + self.epsilon)
        periods = torch.floor((1 - x) * self.n_steps).type(torch.long)
        spikes = to_tensor(self.firing_periods_to_spikes(periods.cpu().numpy()))
        if squeeze_flag:
            spikes = spikes.squeeze()
        return spikes.to(device)


class ConstantValuesTransform(torch.nn.Module):
    def __init__(self, n_steps: int, batch_wise: bool = True):
        super().__init__()
        self.n_steps = n_steps
        self.batch_wise = batch_wise

    def __repr__(self):
        return f"{self.__class__.__name__}(n_steps={self.n_steps})"

    def forward(self, x: Any):
        x = to_tensor(x)
        if self.batch_wise:
            if x.ndim < 3:
                x = x.unsqueeze(1)
            return x.repeat(1, self.n_steps, 1)
        if x.ndim < 2:
            x = x.unsqueeze(0)
        return x.repeat(self.n_steps, 1)


class MaybeSoftmax(torch.nn.Module):
    def __init__(self, dim: int = -1):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        from ..utils import maybe_apply_softmax
        return maybe_apply_softmax(x, self.dim)


class ReduceMax(torch.nn.Module):
    def __init__(self, dim: int = 1):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        from ..utils import unpack_out_hh
        out, hh = unpack_out_hh(x)
        if isinstance(out, torch.Tensor):
            out_max, _ = torch.max(out, dim=self.dim)
        elif isinstance(out, dict):
            out_max = {
                k: torch.max(v, dim=self.dim)[0]
                for k, v in out.items()
            }
        else:
            raise ValueError("Inputs must be a torch.Tensor or a dictionary.")
        return out_max

    def extra_repr(self) -> str:
        return f"dim={self.dim}"


class ReduceMean(torch.nn.Module):
    def __init__(self, dim: int = 1):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        from ..utils import unpack_out_hh
        out, hh = unpack_out_hh(x)
        if isinstance(out, torch.Tensor):
            out_mean = torch.mean(out, dim=self.dim)
        elif isinstance(out, dict):
            out_mean = {
                k: torch.mean(v, dim=self.dim)
                for k, v in out.items()
            }
        else:
            raise ValueError("Inputs must be a torch.Tensor or a dictionary.")
        return out_mean

    def extra_repr(self) -> str:
        return f"dim={self.dim}"


class ReduceSum(torch.nn.Module):
    def __init__(self, dim: int = 1):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        from ..utils import unpack_out_hh
        out, hh = unpack_out_hh(x)
        if isinstance(out, torch.Tensor):
            out_sum = torch.sum(out, dim=self.dim)
        elif isinstance(out, dict):
            out_sum = {
                k: torch.sum(v, dim=self.dim)
                for k, v in out.items()
            }
        else:
            raise ValueError("Inputs must be a torch.Tensor or a dictionary.")
        return out_sum

    def extra_repr(self) -> str:
        return f"dim={self.dim}"


class ReduceFuncTanh(torch.nn.Module):
    """
    Applies a reduction function to the output of a recurrent layer and then applies a tanh activation.
    """
    def __init__(
            self,
            reduce_func: Callable[
                [Union[Dict[str, torch.Tensor], torch.Tensor]],
                Union[Dict[str, torch.Tensor], torch.Tensor]
            ]
    ):
        """
        Constructor of the ReduceFuncTanh class.

        :param reduce_func: The reduction function to apply to the output of the recurrent layer. Must take a
            torch.Tensor or a dictionary of shape(s) (batch_size, seq_len, hidden_size) as input and return a
            torch.Tensor or a dictionary of shape(s) (batch_size, hidden_size) as output.
        """
        super().__init__()
        self.reduce_func = reduce_func
        self.tanh = torch.nn.Tanh()

    def forward(self, x):
        out_reduced = self.reduce_func(x)
        if isinstance(out_reduced, torch.Tensor):
            out_reduced_tanh = self.tanh(out_reduced)
        elif isinstance(out_reduced, dict):
            out_reduced_tanh = {
                k: self.tanh(v)
                for k, v in out_reduced.items()
            }
        else:
            raise ValueError("Inputs must be a torch.Tensor or a dictionary.")
        return out_reduced_tanh




import torch
import torch.nn as nn

from ..transforms.base import to_tensor


class RMSELoss(torch.nn.Module):
    """
    Class used to compute the RMSE loss.

    :math:`\\text{RMSE}(x, y) = \\sqrt{\\frac{1}{n}\\sum_{i=1}^n (x_i - y_i)^2}`

    :Attributes:
        - **criterion** (nn.MSELoss): The MSE loss.

    """
    def __init__(self):
        """
        Constructor for the RMSELoss class.
        """
        super(RMSELoss, self).__init__()
        self.criterion = nn.MSELoss()

    def forward(self, x, y):
        """
        Calculate the RMSE loss.

        :param x: The first input.
        :param y: The second input.

        :return: The RMSE loss.
        """
        loss = self.criterion(x, y)
        loss = torch.pow(loss + 1e-8, 0.5)
        return loss


class PVarianceLoss(torch.nn.Module):
    """
    Class used to compute the P-Variance loss.

    :math:`\\text{P-Variance}(x, y) = 1 - \\frac{\\text{MSE}(x, y)}{\\text{Var}(y)}`

    :Attributes:
        - :attr:`criterion` (nn.MSELoss): The MSE loss.
        - :attr:`negative` (bool): Whether to return the negative P-Variance loss.
        - :attr:`reduction` (str): The reduction method to use. If 'mean', the output will be averaged. If 'feature', the
            output will be the shape of the last dimension of the input. If 'none', the output will be the same shape as
            the input.

    """
    def __init__(self, negative: bool = False, reduction: str = 'mean', **kwargs):
        """
        Constructor for the PVarianceLoss class.

        :param negative: Whether to return the negative P-Variance loss.
        :type negative: bool
        :param reduction: The reduction method to use. If 'mean', the output will be averaged. If 'feature', the output
            will be the shape of the last dimension of the input. If 'none', the output will be the same shape as the
            input. Defaults to 'mean'.
        :type reduction: str
        :keyword arguments : epsilon: The epsilon value to use to prevent division by zero. Defaults to 1e-5.
        """
        super(PVarianceLoss, self).__init__()
        assert reduction in ['mean', 'feature', 'none'], 'Reduction must be one of "mean", "feature", or "none".'
        self.reduction = reduction
        mse_reduction = 'mean' if reduction == 'mean' else 'none'
        self.criterion = nn.MSELoss(
            reduction=mse_reduction
        )
        self.negative = negative
        self.epsilon = kwargs.get("epsilon", 1e-5)

    def forward(self, x, y):
        """
        Calculate the P-Variance loss.

        :param x: The first input.
        :param y: The second input.

        :return: The P-Variance loss.
        """
        x, y = to_tensor(x), to_tensor(y)
        if self.reduction == 'feature':
            x_reshape, y_reshape = x.reshape(-1, x.shape[-1]), y.reshape(-1, y.shape[-1])
        else:
            x_reshape, y_reshape = x, y
        mse_loss = self.criterion(x_reshape, y_reshape)
        if self.reduction == 'feature':
            mse_loss = mse_loss.mean(dim=0)
            var = y_reshape.var(dim=0)
        else:
            var = y_reshape.var()
        loss = 1 - (mse_loss / (var + self.epsilon))
        if self.negative:
            loss = -loss
        return loss

    def mean_std_over_batch(self, x, y):
        """
        Calculate the mean and standard deviation of the P-Variance loss over the batch.

        :param x: The first input.
        :param y: The second input.

        :return: The mean and standard deviation of the P-Variance loss over the batch.
        """
        x, y = to_tensor(x), to_tensor(y)
        x_reshape, y_reshape = x.reshape(x.shape[0], -1), y.reshape(y.shape[0], -1)
        mse_loss = torch.mean((x_reshape - y_reshape)**2, dim=-1)
        var = y_reshape.var(dim=-1)
        loss = 1 - (mse_loss / (var + self.epsilon))
        if self.negative:
            loss = -loss
        return loss.mean(), loss.std()


class NLLLoss(torch.nn.NLLLoss):
    """
    Class used to compute the negative log likelihood loss.

    :math:`\\text{NLLLoss}(x, y) = -\\sum_{i=1}^n y_i \\log(x_i)`

    The difference between this class and the PyTorch NLLLoss class is that this class allows ND inputs and targets
    by flattening the inputs to 2D and the targets to 1D. If the target is marked as one-hot encoded, then the target
    will be converted to a 1D tensor of class indices by taking the argmax of the last dimension.
    """
    def __init__(self, target_as_one_hot: bool = False, **kwargs):
        """
        Constructor for the NLLLoss class.

        :param target_as_one_hot: Whether the target is one-hot encoded. Defaults to False.
        :keyword arguments : Arguments for the PyTorch NLLLoss class. See the PyTorch documentation for more details.
        """
        super(NLLLoss, self).__init__(**kwargs)
        self.target_as_one_hot = target_as_one_hot

    def forward(self, inputs: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Calculate the NLL loss.

        :param inputs: The input.
        :param target: The target.

        :return: The NLL loss.
        """
        if self.target_as_one_hot:
            target = target.argmax(dim=-1)
        inputs_view, target_view = inputs.view(-1, inputs.shape[-1]), target.view(-1)
        loss = super(NLLLoss, self).forward(inputs_view, target_view)
        return loss


class SMSEloss(torch.nn.Module):
    def __init__(self, reduction: str = "mean", **kwargs):
        super().__init__()
        assert reduction in ['mean', 'feature', 'none'], 'Reduction must be one of "mean", "feature", or "none".'
        self.reduction = reduction
        self.epsilon = kwargs.get("epsilon", 1e-5)

    def forward(self, inputs: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        x, y = to_tensor(inputs), to_tensor(target)
        if self.reduction == 'feature':
            x_reshape, y_reshape = x.reshape(-1, x.shape[-1]), y.reshape(-1, y.shape[-1])
        else:
            x_reshape, y_reshape = x, y
        mse_loss = torch.nn.functional.mse_loss(x, y, reduction=self.reduction)
        if self.reduction == 'feature':
            mse_loss = mse_loss.mean(dim=0)
            var = y_reshape.var(dim=0)
        else:
            var = y_reshape.var()
        return mse_loss / (var + self.epsilon)



import os
from typing import Dict, List, Tuple, Optional

import numpy as np
import matplotlib.pyplot as plt

from ..transforms.base import to_numpy
from ..callbacks.base_callback import BaseCallback


class TrainingHistory(BaseCallback):
    """
    This class is used to store some metrics over the training process.

    :Attributes:
        - **default_value** (float): The default value to use to equalize the lengths of the container's items.
    """
    DEFAULT_PRIORITY = BaseCallback.DEFAULT_HIGH_PRIORITY

    @staticmethod
    def _set_default_plot_kwargs(kwargs: dict):
        kwargs.setdefault('fontsize', 18)
        kwargs.setdefault('linewidth', 4)
        kwargs.setdefault('figsize', (16, 12))
        kwargs.setdefault('dpi', 300)
        return kwargs

    @staticmethod
    def _remove_prefix_from_metrics(metrics: List[str]):
        return [metric.split('_')[-1] for metric in metrics]

    def __init__(
            self,
            container: Dict[str, List[float]] = None,
            default_value=np.NAN,
            **kwargs
    ):
        """
        Initialize the container with the given container.

        :param container: The container to initialize the container with.
        :type container: Dict[str, List[float]]
        :param default_value: The default value to use to equalize the lengths of the container's items.
        :type default_value: float
        :param kwargs: The keyword arguments to pass to the BaseCallback.
        """
        super().__init__(**kwargs)
        self._container = {}
        self.default_value = default_value
        self._length = 0
        if container is not None:
            self.concat(container)

    def __getitem__(self, key):
        self._add_key(key)
        return self._container[key]

    def __setitem__(self, key, value: list):
        self._add_key(key)
        assert len(value) == len(self), "Length of value must be equal to length of container"
        self._container[key] = list(value)

    def __contains__(self, item):
        return item in self._container

    def __iter__(self):
        return iter(self._container)

    def __len__(self):
        return self._length

    def _add_key(self, key):
        if key not in self._container:
            self._container[key] = [self.default_value] * len(self)

    def _increase_to_size(self, to_size: int):
        """
        Increase the size of the container items to the given size.

        :param to_size: The size to increase the container to.
        :type to_size: int

        :return: None
        """
        if len(self) > to_size:
            raise ValueError("Cannot increase size of container to smaller size")
        if len(self) == to_size:
            return
        for key, values in self.items():
            self[key].extend([self.default_value] * (to_size - len(self)))
        self._length = to_size

    def keys(self):
        return self._container.keys()

    def items(self):
        return self._container.items()

    def concat(self, other):
        self.insert(len(self), other)

    def insert(self, index: int, other):
        """
        Increase the size of the container items to the given index and insert the given other into the container.

        :param index: The index to insert the other at.
        :type index: int
        :param other: The other to insert.
        :type other: Dict[str, float]

        :return: None
        """
        if index >= len(self):
            self._increase_to_size(index + 1)
        for key, values in other.items():
            if isinstance(values, list):
                raise NotImplementedError()
            else:
                self[key][index] = values

    def append(self, key, value):
        """
        Add the given value to the given key. Increase the size of the container by one.

        :param key: The key to add the value to.
        :type key: str
        :param value: The value to add.
        :type value: float

        :return: None
        """
        self.insert(len(self), {key: value})

    def min(self, key=None, default: float = np.inf) -> float:
        """
        Get the minimum value of the given key.

        :param key: The key to get the minimum value of. If None, the first key is used.
        :type key: str
        :param default: The default value to return if the key is not in the container.
        :type default: float

        :return: The minimum value of the given key.
        :rtype: float
        """
        if key is None:
            key = list(self.keys())[0]
        if key in self and len(self[key]) > 0:
            return np.nanmin(self[key])
        return default

    def min_item(self, key=None) -> Dict[str, float]:
        """
        Get all the metrics of the iteration with the minimum value of the given key.

        :param key: The key to get the minimum value of. If None, the first key is used.
        :type key: str

        :return: All the metrics of the iteration with the minimum value of the given key.
        :rtype: Dict[str, float]

        :raises ValueError: If the key is not in the container.
        """
        if key is None:
            key = list(self.keys())[0]
        if key in self:
            argmin = np.argmin(self[key])
            return {k: v[argmin] for k, v in self.items()}
        raise ValueError("key not in container")

    def max(self, key=None, default=-np.inf):
        """
        Get the maximum value of the given key.

        :param key: The key to get the maximum value of. If None, the first key is used.
        :type key: str
        :param default: The default value to return if the key is not in the container.
        :type default: float

        :return: The maximum value of the given key.
        :rtype: float
        """
        if key is None:
            key = list(self.keys())[0]
        if key in self and len(self[key]) > 0:
            return np.nanmax(self[key])
        return default

    def max_item(self, key=None):
        """
        Get all the metrics of the iteration with the maximum value of the given key.

        :param key: The key to get the maximum value of. If None, the first key is used.
        :type key: str

        :return: All the metrics of the iteration with the maximum value of the given key.
        :rtype: Dict[str, float]

        :raises ValueError: If the key is not in the container.
        """
        if key is None:
            key = list(self.keys())[0]
        if key in self:
            argmax = np.argmax(self[key])
            return {k: v[argmax] for k, v in self.items()}
        raise ValueError("key not in container")

    def get_item_at(self, idx: int = -1):
        """
        Get all the metrics of the iteration at the given index.

        :param idx: The index to get the metrics of.
        :type idx: int

        :return: All the metrics of the iteration at the given index.
        :rtype: Dict[str, float]

        :raises ValueError: If the index is out of bounds.
        """
        if idx >= len(self):
            raise ValueError("Index out of bounds")
        return {k: v[idx] for k, v in self.items()}

    def get(self, key, default=None):
        """
        Get the values of the given key.

        :param key: The key to get the values of.
        :type key: str
        :param default: The default value to return if the key is not in the container.
        :type default: Any

        :return: The values of the given key.
        :rtype: List[float]
        """
        return self._container.get(key, default)

    def create_plot(self, **kwargs) -> Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]:
        """
        Create a plot of the metrics in the container.

        :param kwargs: Keyword arguments.

        :keyword Tuple[float, float] figsize: The size of the figure.
        :keyword int linewidth: The width of the lines.

        :return: The figure, axes and lines of the plot.
        :rtype: Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]
        """
        kwargs = self._set_default_plot_kwargs(kwargs)
        keys_lower_to_given = {k.lower(): k for k in self.keys()}
        keys_lower = [key.lower() for key in self.keys()]
        loss_metrics = [k for k in keys_lower if 'loss' in k]
        keys_lower = list(set(keys_lower) - set(loss_metrics))
        val_metrics = [k for k in keys_lower if 'val' in k]
        train_metrics = [k for k in keys_lower if 'train' in k]
        test_metrics = [k for k in keys_lower if 'test' in k]
        n_set_metrics = max(len(val_metrics), len(train_metrics), len(test_metrics))
        max_set_metrics_container = [c for c in [val_metrics, train_metrics, test_metrics] if len(c) == n_set_metrics][0]
        other_metrics = list(set(keys_lower) - set(val_metrics) - set(train_metrics) - set(test_metrics))
        n_graphs = 1 + n_set_metrics + len(other_metrics)
        n_cols = int(np.sqrt(n_graphs))
        n_rows = int(np.ceil(n_graphs / n_cols))
        axes_dict, lines = {}, {}
        fig, axes = plt.subplots(nrows=n_rows, ncols=n_cols, figsize=kwargs["figsize"], sharex='all')
        axes = np.ravel(axes)
        for i, ax in enumerate(axes):
            if i >= n_graphs:
                ax.axis('off')
                continue
            if i == 0:
                for k in loss_metrics:
                    key = keys_lower_to_given[k]
                    lines[key] = ax.plot(self[key], label=key, linewidth=kwargs['linewidth'])[0]
                axes_dict['losses'] = ax
                ax.set_ylabel("Loss [-]", fontsize=kwargs["fontsize"])
                ax.set_xlabel("Iterations [-]", fontsize=kwargs["fontsize"])
                ax.legend(fontsize=kwargs["fontsize"])
            elif 0 < i <= n_set_metrics:
                metric_basename = '_'.join(max_set_metrics_container[i-1].split('_')[1:])
                for prefix in ['val', 'train', 'test']:
                    k = prefix + '_' + metric_basename
                    if k in keys_lower_to_given:
                        key = keys_lower_to_given[k]
                        if key in self:
                            lines[key] = ax.plot(self[key], label=key, linewidth=kwargs['linewidth'])[0]
                            axes_dict[key] = ax
                ax.set_xlabel("Iterations [-]", fontsize=kwargs["fontsize"])
                ax.legend(fontsize=kwargs["fontsize"])
            else:
                k = other_metrics[i - 1 - n_set_metrics]
                key = keys_lower_to_given[k]
                lines[key] = ax.plot(self[key], label=key, linewidth=kwargs['linewidth'])[0]
                axes_dict[key] = ax
                ax.set_xlabel("Iterations [-]", fontsize=kwargs["fontsize"])
                ax.legend(fontsize=kwargs["fontsize"])
        return fig, axes_dict, lines

    def plot(
            self,
            save_path: Optional[str] = None,
            show: bool = False,
            **kwargs
    ) -> Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]:
        """
        Plot the metrics in the container.

        :param save_path: The path to save the plot to. If None, the plot is not saved.
        :type save_path: Optional[str]

        :param show: Whether to show the plot.
        :type show: bool

        :param kwargs: Keyword arguments.

        :keyword Tuple[float, float] figsize: The size of the figure.
        :keyword int linewidth: The width of the lines.
        :keyword int dpi: The resolution of the figure.
        :keyword bool block: Whether to block execution until the plot is closed.
        :keyword bool close: Whether to close the plot at the end.

        :return: The figure, axes and lines of the plot.
        :rtype: Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]
        """
        kwargs = self._set_default_plot_kwargs(kwargs)
        plt.close('all')
        fig, axes, lines = self.create_plot(**kwargs)
        if save_path is not None:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            fig.savefig(save_path, dpi=kwargs["dpi"])
        if show:
            plt.show(block=kwargs.get('block', True))
        if kwargs.get('close', True):
            plt.close(fig)
        return fig, axes, lines

    def update_fig(
            self,
            fig: plt.Figure,
            axes: Dict[str, plt.Axes],
            lines: Dict[str, plt.Line2D],
            **kwargs
    ) -> Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]:
        """
        Update the plot of the metrics in the container.

        :param fig: The figure to update.
        :type fig: plt.Figure
        :param axes: The axes to update.
        :type axes: Dict[str, plt.Axes]
        :param lines: The lines to update.
        :type lines: Dict[str, plt.Line2D]
        :param kwargs: Keyword arguments.

        :return: The figure, axes and lines of the plot.
        :rtype: Tuple[plt.Figure, Dict[str, plt.Axes], Dict[str, plt.Line2D]]
        """
        kwargs = self._set_default_plot_kwargs(kwargs)
        for k in lines:
            lines[k].set_data(range(len(self[k])), self[k])
        for k in axes:
            axes[k].relim()
            axes[k].autoscale_view()
        fig.canvas.draw()
        fig.canvas.flush_events()
        return fig, axes, lines

    def on_iteration_end(self, trainer, **kwargs):
        """
        Insert the current metrics into the container.

        :param trainer: The trainer.
        :type trainer: Trainer

        :return: None
        """
        self.insert(
            trainer.current_training_state.iteration,
            {k: to_numpy(v) for k, v in trainer.current_training_state.itr_metrics.items()}
        )

    def extra_repr(self):
        return f", n={len(self)}, metrics={list(self.keys())}"










import numbers

import numpy as np

from .base_callback import (
    BaseCallback,
    CallbacksList,
)

from .checkpoints_manager import (
    LoadCheckpointMode,
    CheckpointManager
)

from .history import (
    TrainingHistory,
)

from . import early_stopping


class ForesightTimeStepUpdaterOnTarget(BaseCallback):
    """
    Updates the foresight time step of the model to the length of the target sequence. This is useful for models that
    are trained with a fixed foresight time step, but are evaluated on sequences of different lengths.
    """
    DEFAULT_PRIORITY = BaseCallback.DEFAULT_HIGH_PRIORITY

    def __init__(self, **kwargs):
        """
        Constructor for ForesightTimeStepUpdaterOnTarget class.

        :keyword int time_steps_multiplier: The multiplier to use to determine the time steps. Defaults to 1.
        :keyword bool target_skip_first: Whether to skip the first time step of the target sequence. Defaults to True.
        :keyword int update_val_loss_freq: The frequency at which to update the validation loss. Defaults to 1.
        :keyword float start_intensive_val_at: The fraction of the training epochs at which to start intensive validation.
            An intensive validation is when the validation is performed at each iteration. Defaults to 0.0.
            If the value is an integer, it is interpreted as the number of iterations at which to start intensive
            validation.
        :keyword str hh_memory_size_strategy: The strategy to use to determine the hidden history memory size.
            The available strategies are:
                - "out_memory_size": The hidden history memory size is equal to the output memory size.
                - "foresight_time_steps": The hidden history memory size is equal to the foresight time steps.
                - <Number>: The hidden history memory size is equal to the specified number.
            Defaults to "out_memory_size".
        """
        kwargs["save_state"] = False
        super().__init__(**kwargs)
        self._hh_memory_size_cache = None
        self._out_memory_size_cache = None
        self._foresight_time_steps_cache = None
        self.time_steps_multiplier = kwargs.get("time_steps_multiplier", 1)
        self.update_val_loss_freq = kwargs.get("update_val_loss_freq", 1)
        self.start_intensive_val_at = kwargs.get("start_intensive_val_at", 0.0)
        if isinstance(self.start_intensive_val_at, float):
            self.start_intensive_val_at = np.clip(self.start_intensive_val_at, 0.0, 1.0)
        elif not isinstance(self.start_intensive_val_at, int):
            raise ValueError("start_intensive_val_at must be a float or an integer.")
        self.hh_memory_size_strategy = kwargs.get("hh_memory_size_strategy", "out_memory_size")
        self.val_dataloader = None
        self.kwargs = kwargs

    def start(self, trainer, **kwargs):
        self.val_dataloader = trainer.state.objects.get("val_dataloader", None)

    def get_hh_memory_size_from_y_batch(self, y_batch) -> int:
        if self.hh_memory_size_strategy == "out_memory_size":
            return y_batch.shape[-2] * self.time_steps_multiplier
        elif self.hh_memory_size_strategy == "foresight_time_steps":
            return y_batch.shape[-2]
        elif isinstance(self.hh_memory_size_strategy, numbers.Number):
            return int(self.hh_memory_size_strategy)
        else:
            raise ValueError(f"Invalid hh_memory_size_strategy: {self.hh_memory_size_strategy}")

    def on_batch_begin(self, trainer, **kwargs):
        self._hh_memory_size_cache = trainer.model.hh_memory_size
        self._out_memory_size_cache = trainer.model.out_memory_size
        self._foresight_time_steps_cache = trainer.model.foresight_time_steps
        if not trainer.model.training:
            n_times_steps = trainer.state.y_batch.shape[-2]
            trainer.model.hh_memory_size = self.get_hh_memory_size_from_y_batch(trainer.state.y_batch)
            trainer.model.out_memory_size = int(n_times_steps * self.time_steps_multiplier)
            trainer.model.foresight_time_steps = trainer.model.out_memory_size

    def on_batch_end(self, trainer, **kwargs):
        trainer.model.hh_memory_size = self._hh_memory_size_cache
        trainer.model.out_memory_size = self._out_memory_size_cache
        trainer.model.foresight_time_steps = self._foresight_time_steps_cache

    def on_train_end(self, trainer, **kwargs):
        if isinstance(self.start_intensive_val_at, float):
            at = int(self.start_intensive_val_at * trainer.state.n_iterations)
        else:
            at = self.start_intensive_val_at
        if (trainer.state.iteration >= at) or ((trainer.state.iteration + 1) % self.update_val_loss_freq == 0):
            trainer.state.objects["val_dataloader"] = self.val_dataloader
        else:
            trainer.state.objects["val_dataloader"] = None
	





import enum
import warnings
from typing import Callable, List, Tuple, Union, Iterable

import torch
import torchvision

from .autograd import (
    zero_grad_params,
    dy_dw_local,
    filter_parameters,
    get_contributing_params,
    recursive_detach,
    recursive_detach_,
)
from .formatting import (
    format_pred_batch,
)
from .random import (
    set_seed,
    format_pseudo_rn_seed,
)
from .visualise import (
    legend_without_duplicate_labels_,
)
from .collections import (
    get_meta_name,
    hash_params,
    get_meta_str,
    hash_meta_str,
    save_params,
    get_all_params_combinations,
    list_of_callable_to_sequential,
    sequence_get,
    list_insert_replace_at,
    unpack_out_hh,
    unpack_tuple,
    unpack_singleton_dict,
    maybe_unpack_singleton_dict,
    mapping_update_recursively,
)
from ..transforms.base import to_tensor


class ConnectivityConvention(enum.Enum):
    """
    Enum class to specify the convention used to define the connectivity matrix.

    .. note::
        - ``ItoJ``: The connectivity matrix is defined as ``W[i, j]`` where ``i`` is the index of the pre-synaptic
                      neuron and ``j`` is the index of the post-synaptic neuron.
        - ``JtoI``: The connectivity matrix is defined as ``W[i, j]`` where ``i`` is the index of the post-synaptic
                      neuron and ``j`` is the index of the pre-synaptic neuron.

    .. note::
        The convention ``ItoJ`` is the one used by default in NeuroTorch. In this case, if the convention is not
        specified in a function, method, class or a feature, it is assumed to be ``ItoJ``.
    """
    ItoJ = 0
    JtoI = 1

    @classmethod
    def from_str(cls, convention: str):
        if convention.lower() in ["itoj", "i_to_j", "i-to-j", "i->j"]:
            return cls.ItoJ
        elif convention.lower() in ["jtoi", "j_to_i", "j-to-i", "j->i"]:
            return cls.JtoI
        else:
            raise ValueError(f"Unrecognized convention: {convention}")

    @classmethod
    def from_other(cls, other):
        if isinstance(other, str):
            return cls.from_str(other)
        elif isinstance(other, cls):
            return other
        else:
            raise ValueError(f"Unrecognized convention: {other}")

    def __str__(self):
        if self == self.ItoJ:
            return "i->j"
        elif self == self.JtoI:
            return "j->i"
        else:
            raise ValueError(f"Unrecognized convention: {self}")


def batchwise_temporal_decay(x: torch.Tensor, decay: float = 0.9):
    r"""

    Apply a decay filter to the input tensor along the temporal dimension.

    :param x: Input of shape (batch_size, time_steps, ...).
    :type x: torch.Tensor
    :param decay: Decay factor of the filter.
    :type decay: float

    :return: Filtered input of shape (batch_size, ...).
    """
    batch_size, time_steps, *_ = x.shape
    assert time_steps >= 1

    powers = torch.arange(time_steps, dtype=torch.float32, device=x.device).flip(0)
    weighs = torch.pow(decay, powers)

    x = torch.mul(x, weighs.unsqueeze(0).unsqueeze(-1))
    x = torch.sum(x, dim=1)
    return x


def batchwise_temporal_filter(x: torch.Tensor, decay: float = 0.9):
    r"""

    Apply a low-pass filter to the input tensor along the temporal dimension.

    .. math::
        \begin{equation}\label{eqn:low-pass-filter}
            \mathcal{F}_\alpha\qty(x^t) = \alpha\mathcal{F}_\alpha\qty(x^{t-1}) + x^t.
        \end{equation}
        :label: eqn:low-pass-filter

    :param x: Input of shape (batch_size, time_steps, ...).
    :type x: torch.Tensor
    :param decay: Decay factor of the filter.
    :type decay: float

    :return: Filtered input of shape (batch_size, time_steps, ...).
    """
    warnings.warn(
        "This function is supposed to compute the same result as `batchwise_temporal_recursive_filter` but it "
        "doesn't. Use `batchwise_temporal_recursive_filter` instead.",
        DeprecationWarning
    )
    batch_size, time_steps, *_ = x.shape
    assert time_steps >= 1

    # TODO: check if this is correct
    powers = torch.arange(time_steps, dtype=torch.float32, device=x.device).flip(0)
    weighs = torch.pow(decay, powers)

    y = torch.mul(x, weighs.unsqueeze(0).unsqueeze(-1))
    y = torch.cumsum(y, dim=1)
    return y


def batchwise_temporal_recursive_filter(x, decay: float = 0.9):
    r"""
    Apply a low-pass filter to the input tensor along the temporal dimension recursively.

    .. math::
        \begin{equation}\label{eqn:low-pass-filter}
            \mathcal{F}_\alpha\qty(x^t) = \alpha\mathcal{F}_\alpha\qty(x^{t-1}) + x^t.
        \end{equation}
        :label: eqn:low-pass-filter

    :param x: Input of shape (batch_size, time_steps, ...).
    :type x: torch.Tensor
    :param decay: Decay factor of the filter.
    :type decay: float

    :return: Filtered input of shape (batch_size, time_steps, ...).
    """
    y = to_tensor(x).detach().clone()
    batch_size, time_steps, *_ = x.shape
    assert time_steps >= 1
    fx = 0.0
    for t in range(time_steps):
        fx = decay * fx + y[:, t]
        y[:, t] = fx
    return y


def linear_decay(init_value, min_value, decay_value, current_itr):
    return max(init_value * decay_value**current_itr, min_value)


def ravel_compose_transforms(
        transform: Union[List, Tuple, torchvision.transforms.Compose, Callable, torch.nn.ModuleList]
) -> List[Callable]:
    transforms = []
    if isinstance(transform, torchvision.transforms.Compose):
        for t in transform.transforms:
            transforms.extend(ravel_compose_transforms(t))
    elif isinstance(transform, (List, Tuple)):
        for t in transform:
            transforms.extend(ravel_compose_transforms(t))
    elif isinstance(transform, torch.nn.Module) and isinstance(transform, Iterable):
        for t in transform:
            transforms.extend(ravel_compose_transforms(t))
    elif callable(transform):
        transforms.append(transform)
    else:
        raise ValueError(f"Unsupported transform type: {type(transform)}")
    return transforms


def maybe_apply_softmax(x, dim: int = -1):
    """
    Apply softmax to x if x is not l1 normalized.

    :Note: The input will be cast to tensor bye the transform `to_tensor`.

    :param x: The tensor to apply softmax to.
    :param dim: The dimension to apply softmax to.
    :return: The softmax applied tensor.
    """
    from ..transforms.base import to_tensor
    # from torch.distributions import constraints
    out = to_tensor(x)
    # constraint = constraints.simplex
    all_positive = torch.all(out >= 0)
    dim_sum = torch.sum(out, dim=dim)
    l1_normalized = torch.allclose(dim_sum, torch.ones_like(dim_sum), atol=1e-6)
    # if constraint.check(out):
    if all_positive and l1_normalized:
        # if torch.allclose(out, out / out.sum(dim=dim, keepdim=True), atol=1e-6):
        return out
    else:
        return torch.nn.functional.softmax(out, dim=dim)


def clip_tensors_norm_(
        tensors: Union[torch.Tensor, Iterable[torch.Tensor]],
        max_norm: float,
        norm_type: float = 2.0,
        error_if_nonfinite: bool = False
) -> torch.Tensor:
    r"""Clips norm of an iterable of tensors.

    This function is a clone from torch.nn.utils.clip_grad_norm_ with the difference that it
    works on tensors instead of parameters.

    The norm is computed over all tensors together, as if they were
    concatenated into a single vector.

    Args:
        tensors (Iterable[Tensor] or Tensor): an iterable of Tensors or a
            single Tensor that will have data normalized
        max_norm (float or int): max norm of the data
        norm_type (float or int): type of the used p-norm. Can be ``'inf'`` for
            infinity norm.
        error_if_nonfinite (bool): if True, an error is thrown if the total
            norm of the data from :attr:`parameters` is ``nan``,
            ``inf``, or ``-inf``. Default: False

    Returns:
        Total norm of the tensors (viewed as a single vector).
    """
    if isinstance(tensors, torch.Tensor):
        tensors = [tensors]
    max_norm = float(max_norm)
    norm_type = float(norm_type)
    if len(tensors) == 0:
        return torch.tensor(0.)
    device = tensors[0].device
    if norm_type == torch.inf:
        norms = [t.detach().abs().max().to(device) for t in tensors]
        total_norm = norms[0] if len(norms) == 1 else torch.max(torch.stack(norms))
    else:
        total_norm = torch.norm(torch.stack([torch.norm(t.detach(), norm_type).to(device) for t in tensors]), norm_type)
    if error_if_nonfinite and torch.logical_or(total_norm.isnan(), total_norm.isinf()):
        raise RuntimeError(
            f'The total norm of order {norm_type} for gradients from '
            '`parameters` is non-finite, so it cannot be clipped. To disable '
            'this error and scale the gradients by the non-finite norm anyway, '
            'set `error_if_nonfinite=False`'
        )
    clip_coef = max_norm / (total_norm + 1e-6)
    # Note: multiplying by the clamped coef is redundant when the coef is clamped to 1, but doing so
    # avoids a `if clip_coef < 1:` conditional which can require a CPU <=> device synchronization
    # when the gradients do not reside in CPU memory.
    clip_coef_clamped = torch.clamp(clip_coef, max=1.0)
    for t in tensors:
        t.detach().mul_(clip_coef_clamped.to(t.device))
    return total_norm



from typing import Optional, Any, Dict, Union, List, Callable

import torch
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from ..modules import BaseModel


class BaseMetrics:
    """
    This base class is used to compute metrics for a model.

    :Attributes:
        - **model** (BaseModel): The model to evaluate.
        - **metrics_names** (List[str]): The metrics to compute.
        - **device** (torch.device): The device in which the metrics will be computed.
    """
    METRICS_NAMES_SEP = ' '

    def __init__(
            self,
            model: BaseModel,
            metrics_names: Any = "all",
            device: Optional[torch.device] = None
    ):
        """
        Constructor for the BaseMetrics class.

        :param model: the model to evaluate.
        :type model: BaseModel
        :param metrics_names: the metrics to compute.
        :type metrics_names: Any
        :param device: The device in which the metrics will be computed.
        :type device: Optional[torch.device]
        """
        self.model = model
        self.metrics_names = self._format_metrics_names_(metrics_names)
        self._check_metrics_names_(self.metrics_names)
        self.device = device

    @property
    def metrics_functions(self) -> Dict[str, Callable]:
        """
        Get the metrics functions to compute.

        :return: The metrics functions to compute.
        :rtype: Dict[str, Callable]
        """
        all_metrics_names_to_func = self.get_unwrap_all_metrics_names_to_func()
        return {
            metric_name: all_metrics_names_to_func[metric_name]
            for metric_name in self.metrics_names
        }

    @classmethod
    def _format_metrics_names_(cls, metrics: Any) -> List[str]:
        if isinstance(metrics, str) and metrics.lower() == "all":
            return cls.get_unique_metrics_names()
        if isinstance(metrics, str):
            metrics = metrics.split(cls.METRICS_NAMES_SEP)
        elif isinstance(metrics, list):
            metrics = [metric.strip() for metric in metrics]
        else:
            raise ValueError("metrics must be a string or a list of strings")
        return metrics

    @classmethod
    def _check_metrics_names_(cls, metrics: List[str]) -> None:
        all_metrics_names = cls.get_all_metrics_names()
        assert all([metric in all_metrics_names for metric in metrics]), \
            f"metrics must be in {all_metrics_names}"

    @staticmethod
    def get_all_metrics_names_to_func() -> Dict[str, Callable]:
        """
        Get all the metrics names and their associated functions.

        :raises NotImplementedError: This method must be implemented in the child class.
        """
        raise NotImplementedError()

    @classmethod
    def get_unwrap_all_metrics_names_to_func(cls) -> Dict[str, Callable]:
        """
        Get all the metrics names and their associated functions. The metrics names are unwrapped.

        :return: The metrics names and their associated functions.
        :rtype: Dict[str, Callable]
        """
        return {
            metric_name: metric_func
            for metric_names, metric_func in cls.get_all_metrics_names_to_func().items()
            for metric_name in metric_names.split(cls.METRICS_NAMES_SEP)
        }

    @classmethod
    def get_all_metrics_names(cls) -> List[str]:
        """
        Get all the metrics names.

        :return: The metrics names.
        :rtype: List[str]
        """
        all_metrics_names = []
        for metric_names, _ in cls.get_all_metrics_names_to_func().items():
            all_metrics_names.extend(metric_names.split(cls.METRICS_NAMES_SEP))
        return all_metrics_names

    @classmethod
    def get_unique_metrics_names(cls) -> List[str]:
        """
        Get all the unique metrics names.

        :return: The unique metrics names.
        :rtype: List[str]
        """
        all_metrics_names = []
        for metric_names, _ in cls.get_all_metrics_names_to_func().items():
            all_metrics_names.append(metric_names.split(cls.METRICS_NAMES_SEP)[0])
        return all_metrics_names

    def __call__(
            self,
            data_loader: DataLoader,
            verbose: Union[bool, int] = False
    ) -> Dict[str, Any]:
        """
        Compute the metrics for the given data_loader.

        :param data_loader: The data loader to use to compute the metrics.
        :type data_loader: DataLoader
        :param verbose: 0: no progress bar, 1: single progress bar, 2: progress bar for each metrics
                        True: verbose is set to 1, False: verbose is set to 0.
        :type verbose: Union[bool, int]

        :return: The metrics computed.
        :rtype: Dict[str, Any]
        """
        if isinstance(verbose, bool):
            verbose = 1 if verbose else 0
        assert verbose in [0, 1, 2], "verbose must be 0, 1 or 2"
        output = {}
        p_bar = tqdm(
            enumerate(self.metrics_functions.items()),
            total=len(self.metrics_functions),
            disable=verbose == 0,
            unit="metric",
            position=0,
        )
        self.model.eval()
        for i, (metric_name, metric_func) in p_bar:
            output[metric_name] = metric_func(
                model=self.model,
                dataloader=data_loader,
                device=self.device,
                verbose=verbose == 2,
                desc=metric_name,
                p_bar_position=i + 1
            )
        return output

















import unittest
# import pytest
from neurotorch import Dimension, DimensionProperty, Size


# @pytest.fixture
# def dimension_init(*args, **kwargs) -> Dimension:
# 	return Dimension(*args, **kwargs)
#
#
# @pytest.mark.parametrize(
# 	"dimension_init, expected_attributes",
# 	[
# 		((), (None, DimensionProperty.NONE, DimensionProperty.NONE.name)),
# 	]
# )
# def test_dimension_init(dimension_init, expected_attributes):
# 	assert dimension_init.size == expected_attributes[0]
# 	assert dimension_init.dtype == expected_attributes[1]
# 	assert dimension_init.name == expected_attributes[2]


class TestDimension(unittest.TestCase):
    def test_default_constructor(self):
        dim = Dimension()
        self.assertIs(dim.size, None)
        self.assertEqual(dim.dtype, DimensionProperty.NONE)
        self.assertEqual(dim.name, dim.dtype.name)

    def test_constructor(self):
        dim = Dimension(10)
        self.assertEqual(dim.size, 10)
        self.assertEqual(dim.dtype, DimensionProperty.NONE)
        self.assertEqual(dim.name, dim.dtype.name)

        dim = Dimension(10, DimensionProperty.NONE)
        self.assertEqual(dim.size, 10)
        self.assertEqual(dim.dtype, DimensionProperty.NONE)
        self.assertEqual(dim.name, dim.dtype.name)

        dim = Dimension(10, DimensionProperty.TIME)
        self.assertEqual(dim.size, 10)
        self.assertEqual(dim.dtype, DimensionProperty.TIME)
        self.assertEqual(dim.name, dim.dtype.name)

        dim = Dimension(10, DimensionProperty.SPATIAL)
        self.assertEqual(dim.size, 10)
        self.assertEqual(dim.dtype, DimensionProperty.SPATIAL)
        self.assertEqual(dim.name, dim.dtype.name)

        dim = Dimension(10, DimensionProperty.SPATIAL)
        self.assertEqual(dim.size, 10)
        self.assertEqual(dim.dtype, DimensionProperty.SPATIAL)
        self.assertEqual(dim.name, dim.dtype.name)






from typing import List, Optional, Sequence

import torch

from ..callbacks.base_callback import BaseCallback


class LearningAlgorithm(BaseCallback):
    DEFAULT_PRIORITY = BaseCallback.DEFAULT_MEDIUM_PRIORITY

    def __init__(
            self,
            *,
            params: Optional[Sequence[torch.nn.Parameter]] = None,
            **kwargs
    ):
        """
        Constructor for LearningAlgorithm class.

        :param params: The parameters to optimize. If None, the parameters of the model's trainer will be used.
        :type params: Optional[Sequence[torch.nn.Parameter]]
        """
        super().__init__(**kwargs)
        self.kwargs = kwargs
        if params is None:
            params = []
        else:
            params = list(params)
        self.params: List[torch.nn.Parameter] = params






import pickle
from collections import defaultdict
from copy import deepcopy
from typing import Any, Iterable, List, NamedTuple, Optional, Dict, Iterator

import numpy as np
import torch
from queue import PriorityQueue

from ..transforms.base import to_tensor, ToDevice, to_numpy


class Experience:
    """
    An experience contains the data of one Agent transition.
    - Observation
    - Action
    - Reward
    - Terminal flag
    - Next Observation
    """
    def __init__(
            self,
            obs: Any,
            action: Any,
            reward: float,
            terminal: bool,
            next_obs: Any,
            discounted_reward: Optional[float] = None,
            advantage: Optional[float] = None,
            rewards_horizon: Optional[List[float]] = None,
            others: Optional[dict] = None
    ):
        self.obs = obs
        self.action = action
        self.reward = reward
        self.terminal = terminal
        self.next_obs = next_obs
        self.rewards_horizon = rewards_horizon or []
        self._discounted_reward = discounted_reward
        self._advantage = advantage
        self.others = others or {}

    @property
    def observation(self):
        return self.obs

    @property
    def metrics(self):
        return self.others

    @metrics.setter
    def metrics(self, metrics):
        self.others = metrics

    @property
    def discounted_reward(self) -> float:
        if self._discounted_reward is None:
            return self.reward
        return self._discounted_reward

    @discounted_reward.setter
    def discounted_reward(self, value: float):
        self._discounted_reward = value

    @property
    def advantage(self) -> float:
        if self._advantage is None:
            return self.discounted_reward
        return self._advantage

    @advantage.setter
    def advantage(self, value: float):
        self._advantage = value


class BatchExperience:
    def __init__(
            self,
            batch: List[Experience],
            device: torch.device = torch.device("cpu"),
    ):
        """
        An object that contains a batch of experiences as tensors.

        :param batch: A list of Experience objects.
        :param device: The device to use for the tensors.
        """
        batch = deepcopy(batch)
        self._batch = batch
        self._device = device
        self._to = ToDevice(device=device)

        self.obs: List[torch.Tensor] = self._make_obs_batch(batch)
        self.rewards: torch.Tensor = self._make_rewards_batch(batch)
        self.terminals: torch.Tensor = self._make_terminals_batch(batch)
        self.actions = self._make_actions_batch(batch)
        self.next_obs: List[torch.Tensor] = self._make_next_obs_batch(batch)
        self.others: List[dict] = [ex.others for ex in batch]

    @property
    def device(self):
        return self._device

    @device.setter
    def device(self, value):
        self._device = value
        # self.obs = [obs.to(value) for obs in self.obs]
        # self.next_obs = [next_obs.to(value) for next_obs in self.next_obs]
        # self.rewards = self.rewards.to(value)
        # self.terminals = self.terminals.to(value)
        # self.continuous_actions = self.continuous_actions.to(value)
        # self.discrete_actions = self.discrete_actions.to(value)
        # self.discounted_rewards = self.discounted_rewards.to(value)
        # self.advantages = self.advantages.to(value)
        self._to.device = value

    def __len__(self):
        return self.rewards.shape[0]

    def __getitem__(self, item):
        if isinstance(item, slice):
            return BatchExperience(batch=self._batch[item], device=self.device)
        return BatchExperience(batch=[self._batch[item]], device=self.device)

    def _make_obs_batch(self, batch: List[Experience]) -> List[torch.Tensor]:
        as_dict = isinstance(batch[0].obs, dict)
        if as_dict:
            obs = {
                key: torch.stack([to_tensor(ex.obs[key]) for ex in batch])
                for key in batch[0].obs
            }
            return self._to(obs)
        return self._to(torch.stack([to_tensor(ex.obs) for ex in batch]))

    def _make_next_obs_batch(self, batch: List[Experience]) -> List[torch.Tensor]:
        as_dict = isinstance(batch[0].next_obs, dict)
        if as_dict:
            obs = {
                key: torch.stack([to_tensor(ex.next_obs[key]) for ex in batch])
                for key in batch[0].next_obs
            }
            return self._to(obs)
        return self._to(torch.stack([to_tensor(ex.next_obs) for ex in batch]))

    def _make_rewards_batch(self, batch: List[Experience]) -> torch.Tensor:
        return self._to(torch.stack([to_tensor(ex.reward) for ex in batch]))

    def _make_terminals_batch(self, batch: List[Experience]) -> torch.Tensor:
        return self._to(torch.stack([to_tensor(ex.terminal) for ex in batch]))

    def _make_discounted_rewards_batch(self, batch: List[Experience]) -> torch.Tensor:
        return self._to(torch.stack([to_tensor(ex.discounted_reward) for ex in batch]))

    def _make_advantages_batch(self, batch: List[Experience]) -> torch.Tensor:
        return self._to(torch.stack([to_tensor(ex.advantage) for ex in batch]))

    def _make_actions_batch(self, batch: List[Experience]) -> torch.Tensor:
        as_dict = isinstance(batch[0].action, dict)
        if as_dict:
            action = {
                key: torch.stack([to_tensor(ex.action[key]) for ex in batch])
                for key in batch[0].action
            }
            return self._to(action)
        return self._to(torch.stack([to_tensor(ex.action) for ex in batch]))

    def _make_rewards_horizon_batch(self, batch: List[Experience]):
        return [self._to(to_tensor(ex.rewards_horizon)) for ex in batch]


class Trajectory:
    """
    A trajectory is a list of experiences.
    """
    def __init__(
            self,
            experiences: Optional[List[Experience]] = None,
            gamma: Optional[float] = None,
            **kwargs,
    ):
        self.experiences = experiences if experiences is not None else []
        self._propagated_flag = False
        self.gamma = gamma
        self.rewards_horizon = kwargs.get("rewards_horizon", 1)
        assert self.rewards_horizon > 0, "The rewards horizon must be greater than 0."

    @property
    def terminated(self):
        return self.experiences[-1].terminal if self.experiences else False

    @property
    def terminal(self):
        return self.terminated

    @property
    def _default_gamma(self):
        return self.gamma if self.gamma is not None else 0.99

    @property
    def cumulative_reward(self):
        return sum([exp.reward for i, exp in enumerate(self.experiences)])

    @property
    def terminal_reward(self):
        return self.experiences[-1].reward

    @property
    def propagated(self):
        return self._propagated_flag

    def is_empty(self):
        return len(self) == 0

    def propagate(self):
        self.propagate_rewards()
        self.make_rewards_horizon()
        self._propagated_flag = True

    def propagate_rewards(self, gamma: Optional[float] = 0.99):
        """
        Propagate the rewards to the next experiences.
        """
        gamma = gamma if gamma is not None else self._default_gamma
        for i in reversed(range(len(self.experiences))):
            if i == len(self.experiences) - 1:
                self.experiences[i].discounted_reward = self.experiences[i].reward
            else:
                self.experiences[i].discounted_reward = (
                        self.experiences[i].reward + gamma * self.experiences[i + 1].discounted_reward
                )

    def make_rewards_horizon(self):
        for i in range(len(self.experiences)):
            self.experiences[i].rewards_horizon = [self.experiences[i].reward]
            for j in range(i, min(i + self.rewards_horizon, len(self.experiences))):
                self.experiences[i].rewards_horizon.append(self.experiences[j].reward)

    def compute_horizon_rewards(self):
        raise NotImplementedError()

    def propagate_values(self, lmbda: Optional[float] = 0.95):
        raise NotImplementedError("Not implemented yet.")

    def __iter__(self):
        return iter(self.experiences)

    def __len__(self):
        return len(self.experiences)

    def __getitem__(self, index: int) -> Experience:
        return self.experiences[index]

    def append(self, experience: Experience):
        self.experiences.append(experience)
        self._propagated_flag = False
        if experience.terminal:
            self.propagate()

    def append_and_propagate(self, experience: Experience):
        self.append(experience)
        self.propagate()

    def update_others(self, others_list: List[dict]):
        assert len(others_list) == len(self.experiences), "The number of experiences must be the same."
        for i, others in enumerate(others_list):
            self.experiences[i].others.update(others)


class ReplayBuffer:
    def __init__(self, capacity=np.inf, seed=None, use_priority=False, **kwargs):
        self.__capacity = capacity
        self._seed = seed
        self.random_generator = np.random.RandomState(seed)
        self.data: List[Experience] = []
        self._counter = 0
        self._counter_is_started = False
        self.use_priority = use_priority
        self.priority_key = kwargs.get("priority_key", "discounted_reward")

    @property
    def counter(self):
        return self._counter

    @property
    def capacity(self):
        return self.__capacity

    @property
    def full(self):
        return len(self) >= self.capacity

    @property
    def empty(self):
        return len(self) == 0

    def __str__(self):
        _repr = f"ReplayBuffer("
        _repr += f"capacity={self.capacity}"
        _repr += f", size={len(self)}"
        _repr += f", use_priority={self.use_priority}"
        _repr += f", seed={self._seed}"
        _repr += f")"
        return _repr

    def set_seed(self, seed: int):
        self.random_generator.seed(seed)

    def start_counter(self):
        self._counter_is_started = True
        self._counter = 0

    def stop_counter(self):
        self._counter_is_started = False
        self._counter = 0

    def reset_counter(self):
        self.stop_counter()

    def increment_counter(self, increment: int = 1):
        self._counter += increment

    def increase_capacity(self, increment: int):
        self.__capacity += increment

    def extend(self, iterable: Iterable[Experience]) -> 'ReplayBuffer':
        _ = [self.store(e) for e in iterable]
        return self

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, idx: int) -> Experience:
        return self.data[idx]

    def store(self, element: Experience) -> 'ReplayBuffer':
        """
        Stores an element. If the replay buffer is already full, deletes the oldest
        element to make space.
        """
        if len(self.data) >= self.__capacity:
            if self.use_priority:
                self.data.pop(np.argmin([np.abs(getattr(e, self.priority_key, 0.0)) for e in self.data]))
            else:
                self.data.pop(0)
        self.data.append(deepcopy(element))
        if self._counter_is_started:
            self._counter += 1
        return self

    def get_random_batch(self, batch_size: int) -> List[Experience]:
        """
        Returns a list of batch_size elements from the buffer.
        """
        return self.random_generator.choice(self.data, size=batch_size)

    def get_batch_tensor(self, batch_size: int, device='cpu') -> BatchExperience:
        """
        Returns a list of batch_size elements from the buffer.
        """
        batch = self.get_random_batch(batch_size)
        return BatchExperience(batch, device=device)

    def get_batch_generator(
            self,
            batch_size: int,
            n_batches: int = None,
            randomize: bool = True,
            device='cpu',
    ) -> Iterator[BatchExperience]:
        """
        Returns a generator of batch_size elements from the buffer.
        """
        if batch_size > len(self.data) or batch_size <= 0:
            batch_size = len(self.data)
        max_idx = int(batch_size * int(len(self) / batch_size))
        indexes = np.arange(max_idx)
        if randomize:
            self.random_generator.shuffle(indexes)
        indexes = indexes.reshape(-1, batch_size)
        if n_batches is None or n_batches > len(indexes) or n_batches < 0:
            n_batches = indexes.shape[0]
        else:
            n_batches = min(n_batches, indexes.shape[0])
        for i in range(n_batches):
            batch = [self.data[j] for j in indexes[i]]
            yield BatchExperience(batch, device=device)

    def clear(self):
        self.data.clear()

    def save(self, filename: str):
        buffer_copy = deepcopy(self)
        for i, e in enumerate(buffer_copy.data):
            buffer_copy.data[i] = Experience(
                obs=e.obs,
                action=e.action,
                reward=e.reward,
                terminal=e.terminal,
                next_obs=e.next_obs,
            )
        with open(filename, 'wb') as file:
            pickle.dump(self, file)

    @staticmethod
    def load(filename: str) -> 'ReplayBuffer':
        with open(filename, 'rb') as file:
            buffer = pickle.load(file)
        for i, e in enumerate(buffer.data):
            buffer.data[i] = Experience(
                obs=e.obs,
                action=e.action,
                reward=e.reward,
                terminal=e.terminal,
                next_obs=e.next_obs,
            )
        return buffer


def _re_zero(*args, **kwargs):
    """
    Use instead of 'lambda: 0.0' to avoid pickling issues.
    """
    return 0.0


class AgentsHistoryMaps:
    r"""
    Class to store the mapping between agents and their history maps

    Attributes:
        trajectories (Dict[int, Trajectory]): Mapping between agent ids and their trajectories
        cumulative_rewards (Dict[int, float]): Mapping between agent ids and their cumulative rewards
    """

    def __init__(self, buffer: Optional[ReplayBuffer] = None, **kwargs):
        self.buffer = buffer if buffer is not None else ReplayBuffer()
        self.trajectories: Dict[int, Trajectory] = defaultdict(Trajectory)
        self.trajectories.update(kwargs.get('trajectories', {}))
        self.cumulative_rewards: Dict[int, list] = defaultdict(list)
        self.terminal_rewards: Dict[int, float] = defaultdict(_re_zero)
        self._terminal_counter = 0
        self._experience_counter = 0
        self.min_rewards = kwargs.get("min_rewards", float('inf'))
        self.max_rewards = kwargs.get("max_rewards", float('-inf'))
        self.normalize_rewards = kwargs.get("normalize_rewards", False)

    @property
    def terminals_count(self) -> int:
        """
        :return: The number of terminal steps
        """
        return self._terminal_counter

    @property
    def experience_count(self) -> int:
        """
        :return: The number of experiences
        """
        return self._experience_counter

    @property
    def max_abs_rewards(self) -> float:
        """
        :return: The maximum absolute reward
        """
        return max(abs(self.min_rewards), abs(self.max_rewards))

    @property
    def cumulative_rewards_as_array(self) -> np.ndarray:
        """
        :return: The cumulative rewards as an array
        """
        cum_rewards_list = sum(self.cumulative_rewards.values(), [])
        # cum_rewards_list = []
        # for agent_id, rewards in self.cumulative_rewards.items():
        # 	cum_rewards_list.extend(rewards)
        return np.asarray(cum_rewards_list)

    @property
    def mean_cumulative_rewards(self) -> float:
        """
        :return: The mean cumulative rewards
        """
        cumulative_rewards = self.cumulative_rewards_as_array
        if cumulative_rewards.size == 0:
            return 0.0
        return np.nanmean(cumulative_rewards).item()

    def update_trajectories_(
            self,
            *,
            observations,
            actions,
            next_observations,
            rewards,
            terminals,
            truncated=None,
            infos=None,
            others=None,
    ) -> List[Trajectory]:
        """
        Updates the trajectories of the agents and returns the trajectories of the agents that have been terminated.

        :param observations: The observations
        :param actions: The actions
        :param next_observations: The next observations
        :param rewards: The rewards
        :param terminals: The terminals
        :param truncated: The truncated
        :param infos: The infos
        :param others: The others

        :return: The terminated trajectories.
        """
        from .utils import get_item_from_batch

        actions = deepcopy(to_numpy(actions))
        observations, next_observations = deepcopy(to_numpy(observations)), deepcopy(to_numpy(next_observations))
        rewards, terminals = deepcopy(to_numpy(rewards)), deepcopy(to_numpy(terminals))
        if others is None:
            others = [None] * len(observations)
        self.min_rewards = min(self.min_rewards, np.min(rewards))
        self.max_rewards = max(self.max_rewards, np.max(rewards))
        if self.normalize_rewards:
            rewards = rewards / (self.max_abs_rewards + 1e-8)

        finished_trajectories = []
        for i in range(len(terminals)):
            if self.trajectories[i].terminated:
                continue
            if terminals[i]:
                self.trajectories[i].append_and_propagate(
                    Experience(
                        obs=get_item_from_batch(observations, i),
                        reward=rewards[i],
                        terminal=terminals[i],
                        action=get_item_from_batch(actions, i),
                        next_obs=get_item_from_batch(next_observations, i),
                        others=get_item_from_batch(others, i),
                    )
                )
                self.cumulative_rewards[i].append(self.trajectories[i].cumulative_reward)
                self.terminal_rewards[i] = self.trajectories[i].terminal_reward
                finished_trajectory = self.trajectories.pop(i)
                finished_trajectories.append(finished_trajectory)
                # self.buffer.extend(finished_trajectory)
                self._terminal_counter += 1
                self._experience_counter += 1
            else:
                self.trajectories[i].append(
                    Experience(
                        obs=get_item_from_batch(observations, i),
                        reward=rewards[i],
                        terminal=terminals[i],
                        action=get_item_from_batch(actions, i),
                        next_obs=get_item_from_batch(next_observations, i),
                        others=get_item_from_batch(others, i),
                    )
                )
                self._experience_counter += 1
        return finished_trajectories

    def propagate_all(self) -> List[Trajectory]:
        """
        Propagate all the trajectories and return the finished ones.

        :return: All the trajectories.
        :rtype: List[Trajectory]
        """
        trajectories = []
        for i in range(len(self.trajectories)):
            if not self.trajectories[i].propagated:
                self.trajectories[i].propagate()
            if self.trajectories[i].terminated:
                self.cumulative_rewards[i].append(self.trajectories[i].cumulative_reward)
                trajectory = self.trajectories.pop(i)
                trajectories.append(trajectory)
                self._terminal_counter += 1
            else:
                trajectory = self.trajectories[i]
        # self.buffer.extend(trajectory)
        return trajectories

    def propagate_and_get_all(self) -> List[Trajectory]:
        """
        Propagate all the trajectories and return all the trajectories.

        :return: All the trajectories
        :rtype: List[Trajectory]
        """
        trajectories = []
        for i in range(len(self.trajectories)):
            if not self.trajectories[i].propagated:
                self.trajectories[i].propagate()
            if self.trajectories[i].terminated:
                self.cumulative_rewards[i].append(self.trajectories[i].cumulative_reward)
                trajectory = self.trajectories.pop(i)
                self._terminal_counter += 1
            else:
                trajectory = self.trajectories[i]
            trajectories.append(trajectory)
        # self.buffer.extend(trajectory)
        return trajectories

    def clear(self) -> List[Trajectory]:
        trajectories = self.propagate_and_get_all()
        self.trajectories.clear()
        self.cumulative_rewards.clear()
        self._terminal_counter = 0
        self._experience_counter = 0
        return trajectories



import os
import shutil
import time
import warnings
from collections import defaultdict, deque, OrderedDict
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union, NamedTuple

import gym
import numpy as np
import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from .agent import Agent
from .buffers import ReplayBuffer, Trajectory, Experience, BatchExperience, AgentsHistoryMaps
from .ppo import PPO
from .utils import env_batch_step, env_batch_reset, batch_numpy_actions, env_batch_render
from .. import Trainer, LoadCheckpointMode, to_numpy, TrainingHistory
from ..callbacks.base_callback import BaseCallback, CallbacksList
from ..learning_algorithms.learning_algorithm import LearningAlgorithm
from ..modules import BaseModel
from ..utils import linear_decay


class GenTrajectoriesOutput(NamedTuple):
    buffer: ReplayBuffer
    cumulative_rewards: np.ndarray
    agents_history_maps: AgentsHistoryMaps
    trajectories: Optional[List[Trajectory]] = None


class RLAcademy(Trainer):
    CUM_REWARDS_METRIC_KEY = "cum_rewards"
    TERMINAL_REWARDS_METRIC_KEY = "terminal_rewards"

    def __init__(
            self,
            agent: Agent,
            *,
            predict_method: str = "__call__",
            learning_algorithm: Optional[LearningAlgorithm] = None,
            callbacks: Optional[Union[List[BaseCallback], CallbacksList, BaseCallback]] = None,
            verbose: bool = True,
            **kwargs
    ):
        self.agent = agent
        kwargs = self.set_default_academy_kwargs(**kwargs)
        super().__init__(
            model=agent,
            predict_method=predict_method,
            learning_algorithm=learning_algorithm,
            callbacks=callbacks,
            verbose=verbose,
            **kwargs
        )
        self._agents_history_maps_meta = {}

    @property
    def env(self):
        if "env" not in self.state.objects:
            raise ValueError("The environment is not set.")
        return self.state.objects["env"]

    @property
    def policy(self) -> BaseModel:
        """
        :return: The policy of the academy.
        """
        return self.model

    @staticmethod
    def set_default_academy_kwargs(**kwargs) -> Dict[str, Any]:
        """
        Set default values for the kwargs of the fit method.
        :param kwargs:
            close_env: Whether to close the environment after the training.
            n_epochs: Number of epochs to train each iteration.
            init_lr: Initial learning rate.
            min_lr: Minimum learning rate.
            weight_decay: Weight decay.
            init_epsilon: Initial epsilon. Epsilon is the probability of choosing a random action.
            epsilon_decay: Epsilon decay.
            min_epsilon: Minimum epsilon.
            gamma: Discount factor.
            tau: Target network update rate.
            n_batches: Number of batches to train each iteration.
            batch_size: Batch size.
            update_freq: Number of steps between each update.
            curriculum_strength: Strength of the teacher learning strategy.
        :return:
        """
        kwargs.setdefault("close_env", False)
        kwargs.setdefault("init_epsilon", 0.00)
        kwargs.setdefault("epsilon_decay", 0.995)
        kwargs.setdefault("min_epsilon", 0.0)
        kwargs.setdefault("n_batches", None)
        kwargs.setdefault("tau", 0.0)
        kwargs.setdefault("batch_size", 256)
        kwargs.setdefault("buffer_size", np.inf)
        kwargs.setdefault("clear_buffer", True)
        kwargs.setdefault("randomize_buffer", True)
        kwargs.setdefault("n_new_trajectories", None)
        kwargs.setdefault("n_new_experiences", kwargs["buffer_size"])
        kwargs.setdefault("clip_ratio", 0.2)
        kwargs.setdefault("use_priority_buffer", False)
        kwargs.setdefault("buffer_priority_key", "discounted_reward")
        kwargs.setdefault("normalize_rewards", False)
        kwargs.setdefault("rewards_horizon", 128)
        kwargs.setdefault("last_k_rewards", 10)
        kwargs.setdefault("last_k_rewards_key", RLAcademy.CUM_REWARDS_METRIC_KEY)

        assert kwargs["batch_size"] <= kwargs["buffer_size"]
        assert kwargs["batch_size"] > 0
        return kwargs

    def _maybe_add_learning_algorithm(self, learning_algorithm: Optional[LearningAlgorithm]) -> None:
        if len(self.learning_algorithms) == 0 and learning_algorithm is None:
            learning_algorithm = PPO(optimizer=self.optimizer, criterion=self.criterion)
        if learning_algorithm is not None:
            self.callbacks.append(learning_algorithm)

    def copy_policy(self, requires_grad: bool = False) -> BaseModel:
        """
        Copy the policy to a new instance.

        :return: The copied policy.
        """
        policy_copy = deepcopy(self.policy)
        for param in policy_copy.parameters():
            param.requires_grad = requires_grad
        policy_copy.eval()
        return policy_copy

    def copy_agent(self, requires_grad: bool = False) -> Agent:
        """
        Copy the agent to a new instance.

        :return: The copied agent.
        """
        agent_copy = self.agent.copy(requires_grad=requires_grad)
        agent_copy.policy.eval()
        return agent_copy

    def _update_optimizer_(self):
        self.policy_optimizer = torch.optim.Adam(
            self.policy.parameters(),
            lr=self.kwargs["init_lr"],
            weight_decay=self.kwargs["weight_decay"]
        )

    def _update_agents_history_maps_meta(self, agents_history_maps: AgentsHistoryMaps):
        self._agents_history_maps_meta = {
            "min_rewards": agents_history_maps.min_rewards,
            "max_rewards": agents_history_maps.max_rewards,
            "trajectories": agents_history_maps.trajectories,
        }

    def reset_agents_history_maps_meta(self):
        self._agents_history_maps_meta = {}
        self.update_objects_state_(observations=None, info=None)

    def generate_trajectories(
            self,
            *,
            n_trajectories: Optional[int] = None,
            n_experiences: Optional[int] = None,
            buffer: Optional[ReplayBuffer] = None,
            epsilon: float = 0.0,
            p_bar_position: int = 0,
            verbose: Optional[bool] = None,
            **kwargs
    ) -> GenTrajectoriesOutput:
        """
        Generate trajectories using the current policy. If the policy of the agent is in evaluation mode, the
        actions will be chosen with the argmax method. If the policy is in training mode and a random number is
        generated that is less than epsilon, a random action will be chosen. Otherwise, the action will be chosen
        by a sample considering the policy output.

        :param n_trajectories: Number of trajectories to generate. If not specified, the number of trajectories
            will be calculated based on the number of experiences.
        :type n_trajectories: int
        :param n_experiences: Number of experiences to generate. If not specified, the number of experiences
            will be calculated based on the number of trajectories.
        :type n_experiences: int
        :param buffer: The buffer to store the experiences.
        :type buffer: ReplayBuffer
        :param epsilon: The probability of choosing a random action.
        :type epsilon: float
        :param p_bar_position: The position of the progress bar.
        :type p_bar_position: int
        :param verbose: Whether to show the progress bar.
        :type verbose: bool
        :param kwargs: Additional arguments.

        :keyword gym.Env env: The environment to generate the trajectories. Will update the "env" of the current_state.
        :keyword observation: The initial observation. If not specified, the observation will be get from the
            the objects of the current_state attribute and if not available, the environment will be reset.
        :keyword info: The initial info. If not specified, the info will be get from the objects of the
            current_state attribute and if not available, the environment will be reset.

        :return: The buffer with the generated experiences, the cumulative rewards and the mean of terminal rewards.
        """
        if n_trajectories is None:
            n_trajectories = self.kwargs["n_new_trajectories"]
        if n_experiences is None:
            n_experiences = self.kwargs["n_new_experiences"]
        if buffer is None:
            buffer = self.current_training_state.objects.get(
                "buffer",
                ReplayBuffer(
                    self.kwargs["buffer_size"],
                    use_priority=self.kwargs["use_priority_buffer"],
                    priority_key=self.kwargs["buffer_priority_key"],
                )
            )
            self.update_objects_state_(buffer=buffer)
        if self.kwargs["clear_buffer"]:
            buffer.clear()
        if verbose is None:
            verbose = self.verbose
        if "env" in kwargs:
            if self.state.objects.get("env", None) != kwargs["env"]:
                self.reset_agents_history_maps_meta()
            self.update_objects_state_(env=kwargs["env"])
        render = kwargs.get("render", self.kwargs.get("render", False))
        rendering = [None for _ in range((self.env.num_envs if hasattr(self.env, "num_envs") else 1))]
        re_trajectories = kwargs.get("re_trajectories", False)
        re_trajectories_list = []
        agents_history_maps = AgentsHistoryMaps(
            # buffer,
            normalize_rewards=self.kwargs["normalize_rewards"], **self._agents_history_maps_meta
        )
        # terminal_rewards: List[float] = []
        p_bar = tqdm(
            total=n_experiences if n_trajectories is None else n_trajectories,
            disable=not verbose, desc="Generating Trajectories", position=p_bar_position,
            unit="trajectory" if n_trajectories is not None else "experience",
        )
        observations = kwargs.get("observations", self.current_training_state.objects.get("observations", None))
        info = kwargs.get("info", self.current_training_state.objects.get("info", None))
        if observations is None or info is None:
            observations, info = env_batch_reset(self.env)
        while not self._update_gen_trajectories_break_flag(agents_history_maps, n_trajectories, n_experiences):
            if render:
                rendering = env_batch_render(self.env)
            if not self.agent.training:
                actions = self.agent.get_actions(observations, env=self.env, re_format="argmax", as_numpy=True)
            elif np.random.random() < epsilon:
                actions = self.agent.get_random_actions(env=self.env, re_format="argmax", as_numpy=True)
            else:
                actions = self.agent.get_actions(observations, env=self.env, re_format="sample", as_numpy=True)
            actions = batch_numpy_actions(actions, self.env)
            next_observations, rewards, dones, truncated, info = env_batch_step(self.env, actions)
            terminals = np.logical_or(dones, truncated)
            finished_trajectories = agents_history_maps.update_trajectories_(
                observations=observations,
                actions=actions,
                next_observations=next_observations,
                rewards=rewards,
                terminals=terminals,
                others=[
                    {"render": rendering_item, "info": info_item, "truncated": truncated_item}
                    for rendering_item, info_item, truncated_item in zip(rendering, info, truncated)
                ],
            )
            # terminal_rewards = list(agents_history_maps.terminal_rewards.values())
            if all(terminals):
                finished_trajectories.extend(agents_history_maps.propagate_all())
                next_observations, info = env_batch_reset(self.env)
            self._update_gen_trajectories_finished_trajectories(finished_trajectories, buffer)
            if re_trajectories:
                re_trajectories_list.extend(finished_trajectories)
            if n_trajectories is None:
                p_bar.update(min(len(terminals), max(0, n_experiences - len(terminals))))
            else:
                p_bar.update(min(sum(terminals), max(0, n_trajectories - sum(terminals))))
            p_bar.set_postfix(
                cumulative_reward=f"{agents_history_maps.mean_cumulative_rewards:.3f}",
                # terminal_rewards=f"{np.nanmean(terminal_rewards) if terminal_rewards else 0.0:.3f}",
            )
            observations = next_observations
        self._update_gen_trajectories_finished_trajectories(agents_history_maps.propagate_and_get_all(), buffer)
        self._update_agents_history_maps_meta(agents_history_maps)
        self.update_objects_state_(observations=observations, info=info, buffer=buffer)
        self.update_itr_metrics_state_(
            **{
                self.CUM_REWARDS_METRIC_KEY: agents_history_maps.mean_cumulative_rewards,
                # self.TERMINAL_REWARDS_METRIC_KEY: np.mean(terminal_rewards),
            }
        )
        p_bar.close()
        return GenTrajectoriesOutput(
            buffer=buffer,
            cumulative_rewards=agents_history_maps.cumulative_rewards_as_array,
            agents_history_maps=agents_history_maps,
            trajectories=re_trajectories_list,
        )

    def _update_gen_trajectories_finished_trajectories(
            self,
            finished_trajectories: List[Trajectory],
            buffer: Optional[ReplayBuffer] = None,
    ):
        if buffer is None:
            buffer = self.state.objects.get("buffer", None)
        for trajectory in finished_trajectories:
            if not trajectory.is_empty():
                trajectory_others_list = self.callbacks.on_trajectory_end(self, trajectory)
                if trajectory_others_list is not None:
                    trajectory.update_others(trajectory_others_list)
                if buffer is not None:
                    buffer.extend(trajectory)
        return buffer

    def _update_gen_trajectories_break_flag(
            self,
            agents_history_maps: AgentsHistoryMaps,
            n_trajectories: Optional[int],
            n_experiences: Optional[int],
    ) -> bool:
        if n_trajectories is not None:
            break_flag = agents_history_maps.terminals_count >= n_trajectories
        elif n_experiences is not None:
            break_flag = agents_history_maps.experience_count >= n_experiences
        else:
            break_flag = agents_history_maps.experience_count >= agents_history_maps.buffer.capacity
        return break_flag

    def _init_train_buffer(self) -> ReplayBuffer:
        self.env.reset()
        buffer, *_ = self.generate_trajectories(n_experiences=self.kwargs["batch_size"])
        return buffer

    def train(
            self,
            env,  # TODO: add eval_env
            n_iterations: Optional[int] = None,
            *,
            n_epochs: int = 10,
            load_checkpoint_mode: LoadCheckpointMode = None,
            force_overwrite: bool = False,
            p_bar_position: Optional[int] = None,
            p_bar_leave: Optional[bool] = None,
            **kwargs
    ) -> TrainingHistory:
        self._load_checkpoint_mode = load_checkpoint_mode
        self._force_overwrite = force_overwrite
        self.kwargs.update(kwargs)
        self.update_state_(
            n_iterations=n_iterations,
            n_epochs=n_epochs,
            objects={
                **self.current_training_state.objects,
                **{"env": env, "last_k_rewards": deque(maxlen=self.kwargs["last_k_rewards"])}
            },
        )
        self.sort_callbacks_()
        self.callbacks.start(self)
        self.load_state()
        if self.current_training_state.iteration is None:
            self.update_state_(iteration=0)
        if len(self.training_history) > 0:
            self.update_itr_metrics_state_(**self.training_history.get_item_at(-1))
        else:
            self.update_state_(itr_metrics={})
        env.reset()
        # buffer = self._init_train_buffer()
        # self.update_objects_state_(buffer=buffer)
        p_bar = tqdm(
            initial=self.current_training_state.iteration,
            total=self.current_training_state.n_iterations,
            desc=kwargs.get("desc", "Training"),
            disable=not self.verbose,
            position=p_bar_position,
            unit="itr",
            leave=p_bar_leave
        )

        for i in self._iterations_generator(p_bar):
            self.update_state_(iteration=i)
            self.callbacks.on_iteration_begin(self)
            epsilon = linear_decay(
                self.kwargs["init_epsilon"], self.kwargs["min_epsilon"], self.kwargs["epsilon_decay"], i
            )
            env = self.current_training_state.objects["env"]
            env.reset()
            itr_metrics = self._exec_iteration(env, epsilon=epsilon)
            self.update_itr_metrics_state_(**itr_metrics, epsilon=epsilon)
            postfix = {f"{k}": f"{v:.5e}" for k, v in self.state.itr_metrics.items()}
            postfix.update(self.callbacks.on_pbar_update(self))
            self.callbacks.on_iteration_end(self)
            p_bar.set_postfix(postfix)
            if self.current_training_state.stop_training_flag:
                p_bar.set_postfix(OrderedDict(**{"stop_flag": "True"}, **postfix))
                break

        self.callbacks.close(self)
        p_bar.close()
        if self.kwargs.get("close_env", True):
            env.close()
        return self.training_history

    def _exec_iteration(
            self,
            env: gym.Env,
            buffer: Optional[ReplayBuffer] = None,
            **kwargs,
    ) -> Dict[str, float]:
        if buffer is None:
            buffer = self.current_training_state.objects.get("buffer", None)
        with torch.no_grad():
            torch.cuda.empty_cache()
        metrics = {}

        self.model.train()
        self.callbacks.on_train_begin(self)

        gen_trajectories_out = self.generate_trajectories(
            n_trajectories=self.kwargs["n_new_trajectories"],
            buffer=buffer,
            epsilon=kwargs.get("epsilon", 0.0),
            p_bar_position=1, verbose=False,
        )
        metrics[self.CUM_REWARDS_METRIC_KEY] = np.mean(gen_trajectories_out.cumulative_rewards)
        # metrics[self.TERMINAL_REWARDS_METRIC_KEY] = np.mean(terminal_rewards)
        self.update_state_(batch_is_train=True)
        train_losses = []
        for epoch_idx in range(self.current_training_state.n_epochs):
            self.update_state_(epoch=epoch_idx)
            train_losses.append(self._exec_epoch(gen_trajectories_out.buffer))
        train_loss = np.mean(train_losses)
        self.update_state_(train_loss=train_loss)
        self.callbacks.on_train_end(self)
        metrics["train_loss"] = train_loss

        # if val_dataloader is not None:
        # 	with torch.no_grad():
        # 		self.model.eval()
        # 		self.callbacks.on_validation_begin(self)
        # 		self.update_state_(batch_is_train=False)
        # 		val_loss = self._exec_epoch(val_dataloader)
        # 		self.update_state_(val_loss=val_loss)
        # 		self.callbacks.on_validation_end(self)
        # 		losses["val_loss"] = val_loss

        last_k_rewards = self.state.objects.get("last_k_rewards", deque(maxlen=self.kwargs["last_k_rewards"]))
        assert self.kwargs["last_k_rewards_key"] in metrics, \
            f"last_k_rewards_key {self.kwargs['last_k_rewards_key']} not in metrics. " \
            f"Please select one of {metrics.keys()}"
        last_k_rewards.append(metrics[self.kwargs["last_k_rewards_key"]])
        metrics[f"mean_last_{self.kwargs['last_k_rewards']}_rewards"] = np.nanmean(last_k_rewards)
        with torch.no_grad():
            torch.cuda.empty_cache()
        return metrics

    def _exec_epoch(
            self,
            buffer: ReplayBuffer,
    ) -> float:
        self.callbacks.on_epoch_begin(self)
        batch_size = min(len(buffer), self.kwargs["batch_size"])
        batches = buffer.get_batch_generator(
            batch_size, self.kwargs["n_batches"],
            randomize=self.kwargs["randomize_buffer"],
            device=self.agent.policy.device,
        )
        batch_losses = []
        for i, exp_batch in enumerate(batches):
            self.update_state_(batch=i)
            batch_losses.append(to_numpy(self._exec_batch(exp_batch)))
        mean_loss = np.mean(batch_losses)
        self.callbacks.on_epoch_end(self)
        return mean_loss

    def _exec_batch(
            self,
            exp_batch: BatchExperience,
            **kwargs,
    ):
        self.update_state_(x_batch=exp_batch)
        self.callbacks.on_batch_begin(self)
        if self.model.training:
            self.callbacks.on_optimization_begin(self, x=exp_batch)
            self.callbacks.on_optimization_end(self)
        else:
            self.callbacks.on_validation_batch_begin(self, x=exp_batch)
            self.callbacks.on_validation_batch_end(self)
        self.callbacks.on_batch_end(self)
        batch_loss = self.current_training_state.batch_loss
        if batch_loss is None:
            batch_loss = 0.0
        elif hasattr(batch_loss, "item") and callable(batch_loss.item):
            batch_loss = batch_loss.item()
        return batch_loss

    def close(self):
        self.env.close()



import numpy as np
from matplotlib import pyplot as plt


def plot_confusion_matrix(cm, classes, ):
    import matplotlib.pyplot as plt
    import itertools

    plt.imshow(cm, interpolation='nearest', cmap=plt.cm.Blues)
    plt.title('Confusion matrix')
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    fmt = 'd'
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        plt.text(
            j, i,
            format(cm[i, j], fmt),
            horizontalalignment="center",
            color="white" if cm[i, j] > thresh else "black"
        )

    plt.ylabel('True label')
    plt.xlabel('Predicted label')
    plt.tight_layout()
    plt.show()


def legend_without_duplicate_labels_(ax: plt.Axes):
    handles, labels = ax.get_legend_handles_labels()
    unique = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
    ax.legend(*zip(*unique))



from typing import Dict, Iterable, Optional, Union
import pythonbasictools as pybt
import torch

from ..callbacks.base_callback import BaseCallback
from ..utils import filter_parameters


class BaseRegularization(torch.nn.Module, BaseCallback):
    """
    Base class for regularization.

    :Attributes:
        - :attr:`params` (torch.nn.ParameterList): The parameters which are regularized.
        - :attr:`Lambda` (float): The weight of the regularization. In other words, the coefficient that multiplies the loss.

    """
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            Lambda: float = 1.0,
            optimizer: Optional[torch.optim.Optimizer] = None,
            **kwargs
    ):
        """
        Constructor of the BaseRegularization class.

        :param params: The parameters which are regularized.
        :type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
        :param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
        :type Lambda: float
        """
        super().__init__()
        BaseCallback.__init__(self, **kwargs)
        if isinstance(params, dict):
            self.params = list(params.values())
        else:
            self.params = list(params)
        self.params = torch.nn.ParameterList(self.params)
        self.Lambda = Lambda
        self.name = self.__class__.__name__
        self.optimizer = optimizer

    def __call__(self, *args, **kwargs) -> torch.Tensor:
        """
        Call the forward pass of the regularization and scale it by the :attr:`Lambda` attribute.

        :param args: args of the forward pass.
        :param kwargs: kwargs of the forward pass.

        :return: The loss of the regularization.
        :rtype: torch.Tensor
        """
        out = super(BaseRegularization, self).__call__(*args, **kwargs)
        return self.Lambda * out

    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        Compute the forward pass of the regularization.

        :param args: args of the forward pass.
        :param kwargs: kwargs of the forward pass.

        :return: The loss of the regularization.
        :rtype: torch.Tensor
        """
        raise NotImplementedError("forward method must be implemented")

    def on_optimization_end(self, trainer, **kwargs):
        reg_loss = self()
        if self.optimizer is not None:
            self.optimizer.zero_grad()
        reg_loss.backward()
        if self.optimizer is not None:
            self.optimizer.step()


class RegularizationList(BaseRegularization):
    """
    Regularization that applies a list of regularization.

    :Attributes:
        - :attr:`regularizations` (Iterable[BaseRegularization]): The regularizations to apply.
    """
    def __init__(
            self,
            regularizations: Optional[Iterable[BaseRegularization]] = None,
            optimizer: Optional[torch.optim.Optimizer] = None,
            **kwargs
    ):
        """
        Constructor of the RegularizationList class.

        :param regularizations: The regularizations to apply.
        :type regularizations: Optional[Iterable[BaseRegularization]]
        """
        self.regularizations = regularizations if regularizations is not None else []
        _params = []
        for regularization in self.regularizations:
            _params.extend(regularization.params)
        super(RegularizationList, self).__init__(
            params=_params,
            Lambda=1.0,
            optimizer=optimizer,
            **kwargs
        )
        self.regularizations = regularizations if regularizations is not None else []

    def __iter__(self):
        """
        Iterate over the regularizations.

        :return: An iterator over the regularizations.
        """
        return iter(self.regularizations)

    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        Compute the forward pass of the regularization.

        :param args: args of the forward pass.
        :param kwargs: kwargs of the forward pass.
        """
        if len(self.regularizations) == 0:
            return torch.tensor(0)
        loss = sum([regularization(*args, **kwargs) for regularization in self.regularizations])
        return loss


# @pybt.docstring.inherit_fields_docstring(fields=["Attributes"], bases=[BaseRegularization])
class Lp(BaseRegularization):
    """
    Regularization that applies LP norm.

    :Attributes:
        - :attr:`p` (int): The p parameter of the LP norm. Example: p=1 -> L1 norm, p=2 -> L2 norm.

    :Note: 0D parameters are not regularized.
    """
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            Lambda: float = 1.0,
            p: int = 1,
            **kwargs
    ):
        """
        Constructor of the L1 class.

        :param params: The parameters which are regularized.
        :type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
        :param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
        :type Lambda: float
        :param p: The p parameter of the LP norm. Example: p=1 -> L1 norm, p=2 -> L2 norm.
        :type p: int
        """
        super(Lp, self).__init__(params, Lambda, **kwargs)
        self.p = p

    def forward(self, *args, **kwargs) -> torch.Tensor:
        """
        Compute the forward pass of the regularization.

        :param args: args of the forward pass
        :param kwargs: kwargs of the forward pass

        :return: The loss of the regularization.
        :rtype: torch.Tensor
        """
        if len(self.params) == 0:
            return torch.tensor(0)
        losses = []
        device = self.params[0].device
        for param in filter_parameters(self.params, requires_grad=True):
            if param.dim() > 0:
                losses.append(torch.linalg.norm(param, self.p).to(device))
        return sum(losses)


# @pybt.docstring.inherit_fields_docstring(fields=["Attributes"], bases=[Lp])
class L1(Lp):
    """
    Regularization that applies L1 norm.
    """
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            Lambda: float = 1.0,
            **kwargs
    ):
        """
        Constructor of the L1 class.

        :param params: The parameters which are regularized.
        :type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
        :param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
        :type Lambda: float
        """
        super(L1, self).__init__(params, Lambda, p=1, **kwargs)


# @pybt.docstring.inherit_fields_docstring(fields=["Attributes"], bases=[Lp])
class L2(Lp):
    """
    Regularization that applies L2 norm.
    """
    def __init__(
            self,
            params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]],
            Lambda: float = 1.0,
            **kwargs
    ):
        """
        Constructor of the L2 class.

        :param params: The parameters which are regularized.
        :type params: Union[Iterable[torch.nn.Parameter], Dict[str, torch.nn.Parameter]]
        :param Lambda: The weight of the regularization. In other words, the coefficient that multiplies the loss.
        :type Lambda: float
        """
        super(L2, self).__init__(params, Lambda, p=2, **kwargs)


